// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

/*
This is an example on how the Azure Go library can be used to interact with
the Windows Azure Service.
Note that this is a provided only as an example and that real code should
probably do something more sensible with errors than ignoring them or panicking.
*/
package main

import (
    "flag"
    "fmt"
    "launchpad.net/gwacl"
    "math/rand"
    "os"
    "time"
)

var certFile string
var subscriptionID string
var verbose bool

func getParams() error {
    flag.StringVar(&certFile, "cert", "", "Name of your management certificate file (in PEM format).")
    flag.StringVar(&subscriptionID, "subscriptionid", "", "Your Azure subscription ID.")
    flag.BoolVar(&verbose, "verbose", false, "Print debugging output")

    flag.Parse()

    if certFile == "" {
        return fmt.Errorf("No .pem certificate specified.  Use the -cert option.")
    }
    if subscriptionID == "" {
        return fmt.Errorf("No subscription ID specified.  Use the -subscriptionid option.")
    }
    return nil
}

func checkError(err error) {
    if err != nil {
        panic(err)
    }
}

// makeRandomIdentifier creates an arbitrary identifier of the given length,
// consisting of only ASCII digits and lower-case ASCII letters.
// The identifier will start with the given prefix.  The prefix must be no
// longer than the specified length, or there'll be trouble.
func makeRandomIdentifier(prefix string, length int) string {
    // Only digits and lower-case ASCII letters are accepted.
    const chars = "abcdefghijklmnopqrstuvwxyz0123456789"

    if len(prefix) > length {
        panic(fmt.Errorf("prefix '%s' is more than the requested %d characters long", prefix, length))
    }

    id := prefix
    for len(id) < length {
        id += string(chars[rand.Intn(len(chars))])
    }
    return id
}

func main() {
    rand.Seed(int64(time.Now().Nanosecond()))

    err := getParams()
    if err != nil {
        fmt.Println(err)
        os.Exit(1)
    }

    api, err := gwacl.NewManagementAPI(subscriptionID, certFile)
    gwacl.SetVerbose(verbose)
    checkError(err)

    // Exercise the API.

    // Exercise hosted services API.
    ExerciseHostedServicesAPI(api)

    fmt.Println("All done.")
}

func ExerciseHostedServicesAPI(api *gwacl.ManagementAPI) {
    // A hosted-service name must be between 3 and 24 characters long.
    hostServiceName := makeRandomIdentifier("gwacl", 24)
    fmt.Printf("Creating a hosted service: '%s'...\n", hostServiceName)
    createHostedService := gwacl.NewCreateHostedServiceWithLocation(hostServiceName, "testLabel", "East US")
    err := api.AddHostedService(createHostedService)
    checkError(err)
    fmt.Println("Done creating a hosted service\n")

    fmt.Println("Listing hosted services...")
    hostedServices, err := api.ListHostedServiceDescriptors()
    checkError(err)
    fmt.Printf("Got %d hosted service(s)\n", len(hostedServices))
    if len(hostedServices) > 0 {
        hostedService := hostedServices[0]
        detailedHostedService, err := api.LoadHostedService(hostedService.ServiceName)
        checkError(err)
        fmt.Printf("Hosted service '%s' contains %d deployments\n", hostedService.ServiceName, len(detailedHostedService.Deployments))
    }
    fmt.Println("Done listing hosted services\n")

    fmt.Println("Adding VM deployment...")
    hostname := makeRandomIdentifier("gwaclhost", 64)
    password := makeRandomIdentifier("pwd-", 16)
    vhdName := makeRandomIdentifier("gwacldisk", 16)
    configurationSet := gwacl.NewLinuxProvisioningConfiguration(hostname, "test", password, false)

    storageAccount := makeRandomIdentifier("gwacl", 24)
    storageLabel := makeRandomIdentifier("gwacl", 64)
    fmt.Printf("Requesting storage account with name '%s' and label '%s'...\n", storageAccount, storageLabel)
    cssi := gwacl.NewCreateStorageServiceInputWithLocation(storageAccount, storageLabel, "East US")
    err = api.AddStorageAccount(cssi)
    checkError(err)
    fmt.Println("Done requesting storage account\n")

    mediaLink := fmt.Sprintf("http://%s.blob.core.windows.net/vhds/%s.vhd", storageAccount, vhdName)
    sourceImageName := "b39f27a8b8c64d52b05eac6a62ebad85__Ubuntu-12_04_1-LTS-amd64-server-20121218-en-us-30GB"
    diskName := makeRandomIdentifier("gwacldisk", 16)
    diskLabel := makeRandomIdentifier("gwacl", 64)
    OSVirtualHardDisk := gwacl.NewOSVirtualHardDisk("", diskLabel, diskName, mediaLink, sourceImageName)
    roleName := makeRandomIdentifier("gwaclrole", 16)
    role := gwacl.NewRole("ExtraSmall", roleName, []gwacl.LinuxProvisioningConfiguration{*configurationSet}, OSVirtualHardDisk)
    machineName := makeRandomIdentifier("gwaclmachine", 20)
    deployment := gwacl.NewDeploymentForCreateVMDeployment(machineName, "Staging", machineName, []gwacl.Role{*role}, "")
    err = api.AddDeployment(deployment, hostServiceName)
    checkError(err)
    fmt.Println("Done adding VM deployment\n")

    fmt.Println("Starting VM...")
    err = api.PerformRoleOperation(hostServiceName, deployment.Name, role.RoleName, gwacl.StartRoleOperation)
    checkError(err)
    fmt.Println("Done starting VM\n")

    fmt.Println("Stopping VM...")
    err = api.PerformRoleOperation(hostServiceName, deployment.Name, role.RoleName, gwacl.ShutdownRoleOperation)
    checkError(err)
    fmt.Println("Done stopping VM\n")

    fmt.Println("Deleting deployment...")
    err = api.DeleteDeployment(hostServiceName, deployment.Name)
    checkError(err)
    fmt.Println("Done deleting deployment\n")

    // TODO: it takes some time for a disk previously attached to a deleted VM 
    // to become "not in use" and thus be available for deletion.
    time.Sleep(3 * time.Minute)

    fmt.Println("Deleting VM disk...")
    err = api.DeleteDisk(diskName)
    checkError(err)
    fmt.Println("Done deleting VM disk\n")

    fmt.Println("Deleting hosted service...")
    api.DeleteHostedService(hostServiceName)
    fmt.Println("Done deleting hosted service\n")

    fmt.Printf("Deleting storage account %s...\n", storageAccount)
    api.DeleteStorageAccount(storageAccount)
    fmt.Println("Done deleting storage account\n")
}
