// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

// This is an example of how to use GWACL to interact with the Azure storage
// services API.
//
// Note that it is provided "as-is" and contains very little error handling.
// Real code should handle errors.

package main

import (
    "flag"
    "fmt"
    "io/ioutil"
    "launchpad.net/gwacl"
    "os"
)

var account string
var key string
var filename string
var container string

func checkContainerAndFilename() error {
    if container == "" || filename == "" {
        return fmt.Errorf("Must supply container and filename")
    }
    return nil
}

func badOperationError() error {
    return fmt.Errorf("Must specify one of 'listcontainers', 'list', 'get', 'add', 'delete'")
}

func getParams() (string, error) {
    flag.StringVar(&account, "account", "", "Storage account name")
    flag.StringVar(&key, "key", "", "A valid storage account key (base64 encoded)")
    flag.StringVar(&container, "container", "", "Name of the container to use")
    flag.StringVar(&filename, "file", "", "File containing blob to upload/download")
    flag.Parse()

    if account == "" || key == "" {
        return "", fmt.Errorf("Must supply account and key parameters")
    }

    if len(flag.Args()) != 1 {
        return "", badOperationError()
    }

    switch flag.Arg(0) {
    case "listcontainers":
        return "listcontainers", nil
    case "list":
        if container == "" {
            return "", fmt.Errorf("Must supply container with 'list'")
        }
        return "list", nil
    case "get":
        if container == "" || filename == "" {
            return "", fmt.Errorf("Must supply container and filename with 'list'")
        }
        return "get", nil
    case "add":
        err := checkContainerAndFilename()
        if err != nil {
            return "", err
        }
        return "add", nil
    case "delete":
        err := checkContainerAndFilename()
        if err != nil {
            return "", err
        }
        return "delete", nil
    }
    return "", badOperationError()
}

var Usage = func() {
    fmt.Fprintf(os.Stderr, "%s [args] <listcontainers|list|get|add|delete>\n", os.Args[0])
    flag.PrintDefaults()
}

func dumpError(err error) {
    if err != nil {
        fmt.Fprintf(os.Stderr, "ERROR:")
        fmt.Fprintf(os.Stderr, "%s\n", err)
    }
}

func listcontainers(storage gwacl.StorageContext) {
    res, e := storage.ListContainers()
    if e != nil {
        dumpError(e)
        return
    }
    for _, c := range res.Containers {
        // TODO: embellish with the other container data
        fmt.Println(c.Name)
    }
}

func list(storage gwacl.StorageContext) {
    res, err := storage.ListBlobs(container)
    if err != nil {
        dumpError(err)
        return
    }
    for _, b := range res.Blobs {
        fmt.Printf("%s, %s, %s\n", b.ContentLength, b.LastModified, b.Name)
    }
}

func add(storage gwacl.StorageContext) {
    var err error
    file, err := os.Open(filename)
    if err != nil {
        dumpError(err)
        return
    }
    defer file.Close()
    // NOTE: if the file is larger than 64MB this will fail, it needs blocking.
    err = storage.PutBlock(container, filename, filename, file)
    if err != nil {
        dumpError(err)
        return
    }
    bl := &gwacl.BlockList{}
    bl.Add(gwacl.BlockListLatest, filename)
    err = storage.PutBlockList(container, filename, bl)
    dumpError(err)
}

func deleteblob(storage gwacl.StorageContext) {
    err := storage.DeleteBlob(container, filename)
    dumpError(err)
}

func get(storage gwacl.StorageContext) {
    var err error
    file, err := storage.GetBlob(container, filename)
    if err != nil {
        dumpError(err)
        return
    }
    data, err := ioutil.ReadAll(file)
    if err != nil {
        dumpError(err)
        return
    }
    os.Stdout.Write(data)
}

func main() {
    flag.Usage = Usage
    var err error
    op, err := getParams()
    if err != nil {
        fmt.Fprintf(os.Stderr, "%s\n", err.Error())
        flag.Usage()
        os.Exit(1)
    }

    storage := gwacl.StorageContext{
        Account: account,
        Key:     key,
    }

    switch op {
    case "listcontainers":
        listcontainers(storage)
    case "list":
        list(storage)
    case "add":
        add(storage)
    case "delete":
        deleteblob(storage)
    case "get":
        get(storage)
    }
}
