package gwacl

import (
    "encoding/xml"
    "errors"
    "fmt"
    "net/http"
)

// HTTPError is an extended version of the standard "error" interface.  It
// adds an HTTP status code.
type HTTPError interface {
    error
    StatusCode() int
}

// HTTPStatus is an HTTP status code.
type HTTPStatus int

// Status returns the HTTP status code as an int.
func (s HTTPStatus) StatusCode() int {
    return int(s)
}

// AzureError is an HTTPError returned by the Azure API.  It contains an
// error message, and an Azure-defined error code.
type AzureError struct {
    error      `xml:"-"`
    HTTPStatus `xml:"-"`
    Code       string `xml:"Code"`
    Message    string `xml:"Message"`
}

// *AzureError implements HTTPError.
var _ HTTPError = &AzureError{}

func (e *AzureError) Error() string {
    description := e.error.Error()
    status := e.StatusCode()
    name := http.StatusText(status)
    return fmt.Sprintf("%s: %s - %s (http code %d: %s)", description, e.Code, e.Message, status, name)
}

// ServerError is a generic HTTPError, without any further helpful information
// from the server that we can count on.
type ServerError struct {
    error
    HTTPStatus
}

// *ServerError implements HTTPError.
var _ HTTPError = &ServerError{}

func (e *ServerError) Error() string {
    description := e.error.Error()
    status := e.StatusCode()
    name := http.StatusText(status)
    return fmt.Sprintf("%s (%d: %s)", description, status, name)
}

// NewHTTPError returns the appropriate HTTPError implementation for a given
// HTTP response.
// It takes a status code and response body, rather than just a standard
// http.Response object, because it must work with go-curl as well as Go's
// standard http package.
func newHTTPError(status int, body []byte, description string) HTTPError {
    httpStatus := HTTPStatus(status)
    baseErr := errors.New(description)
    azureError := AzureError{error: baseErr, HTTPStatus: httpStatus}
    err := xml.Unmarshal(body, &azureError)
    if err != nil {
        // It's OK if the response body wasn't actually XML...  That just means
        // it wasn't a proper AzureError.  We have another error type for that.
        return &ServerError{error: baseErr, HTTPStatus: httpStatus}
    }
    return &azureError
}
