// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "encoding/xml"
    "fmt"
    "net/http"
    "time"
)

type ManagementAPI struct {
    session *x509Session
}

// NewManagementAPI creates an object used to interact with Windows Azure's API.
// http://msdn.microsoft.com/en-us/library/windowsazure/ff800682.aspx
func NewManagementAPI(subscriptionId string, certFile string) (*ManagementAPI, error) {
    session, err := newX509Session(subscriptionId, certFile)
    if err != nil {
        return nil, err
    }
    return &ManagementAPI{session}, nil
}

var operationIDHeaderName = http.CanonicalHeaderKey("x-ms-request-id")

// getOperationID extracts the Windows Azure operation ID from the headers
// of the given x509Response.
func getOperationID(response *x509Response) (string, error) {
    header := response.Header[operationIDHeaderName]
    if header != nil && len(header) != 0 {
        return header[0], nil
    }
    err := fmt.Errorf("no operation header (%v) found in response", operationIDHeaderName)
    return "", err
}

// The interval used when polling the server to get the status of a running operation.
// Set this to 0 to prevent polling from happening.
var pollerInterval = 10 * time.Second

// blockUntilCompleted blocks until the operation referenced in the headers of the given response
// is completed.  It returns an error if the given response does not contain
// a reference to an asychronous operation, if the response from the server
// cannot be parsed, or if the asynchronous operation fails.
// Internally, it polls the server at regular intervals until the operation is completed.
func (api *ManagementAPI) blockUntilCompleted(response *x509Response) error {
    if pollerInterval != 0 {
        operationID, err := getOperationID(response)
        if err != nil {
            return err
        }
        poller := newOperationPoller(api, operationID)
        operation, err := performOperationPolling(poller, pollerInterval)
        if err != nil {
            return err
        }
        if operation.Status != SucceededOperationStatus {
            err := fmt.Errorf("asychronous operation failed (status: %v)", operation.Status)
            return err
        }
    }
    return nil
}

// ListHostedServiceDescriptors loads a list of HostedServiceDescriptor objects from the
// Azure management API. 
// HostedServiceDescriptor objects contains a small subset of the fields present in
// HostedService objects.
func (api *ManagementAPI) ListHostedServiceDescriptors() ([]HostedServiceDescriptor, error) {
    URI := "services/hostedservices"
    res, err := api.session.get(URI)
    if err != nil {
        return nil, err
    }
    hostedServices := HostedServiceDescriptorList{}
    err = xml.Unmarshal(res.Body, &hostedServices)
    return hostedServices.HostedServices, err
}

// LoadHostedService loads a HostedService object from the Azure management API.
func (api *ManagementAPI) LoadHostedService(serviceName string) (*HostedService, error) {
    URI := "services/hostedservices/" + serviceName + "?embed-detail=true"
    res, err := api.session.get(URI)
    if err != nil {
        return nil, err
    }
    hostedService := HostedService{}
    err = xml.Unmarshal(res.Body, &hostedService)
    return &hostedService, err
}

// AddHostedService adds a hosted service.
// This is an asynchronous operation, it blocks until the operation is completed.
func (api *ManagementAPI) AddHostedService(definition *CreateHostedService) error {
    URI := "services/hostedservices"
    body, err := marshalXML(definition)
    if err != nil {
        return err
    }
    response, err := api.session.post(URI, []byte(body), "application/xml")
    if err != nil {
        return err
    }
    return api.blockUntilCompleted(response)
}

func (api *ManagementAPI) DeleteHostedService(serviceName string) error {
    _, err := api.session.delete("services/hostedservices/" + serviceName)
    return err
}

// AddDeployment adds a virtual machine deployment.
// This is an asynchronous operation, it blocks until the operation is completed.
func (api *ManagementAPI) AddDeployment(definition *Deployment, serviceName string) error {
    URI := "services/hostedservices/" + serviceName + "/deployments"
    body, err := marshalXML(definition)
    if err != nil {
        return err
    }
    response, err := api.session.post(URI, []byte(body), "application/xml")
    if err != nil {
        return err
    }
    return api.blockUntilCompleted(response)
}

func (api *ManagementAPI) DeleteDeployment(serviceName string, deploymentName string) error {
    path := "services/hostedservices/" + serviceName + "/deployments/" + deploymentName
    _, err := api.session.delete(path)
    return err
}

// AddStorageAccount starts the creation of a storage account.  This is
// called a storage service in the Azure API, but nomenclature seems to
// have changed.
// This is an asynchronous operation, it blocks until the operation is completed.
func (api *ManagementAPI) AddStorageAccount(definition *CreateStorageServiceInput) error {
    uri := "services/storageservices"
    body, err := marshalXML(definition)
    if err != nil {
        return err
    }
    response, err := api.session.post(uri, []byte(body), "application/xml")
    if err != nil {
        return err
    }
    return api.blockUntilCompleted(response)
}

// DeleteStorageAccount deletes a storage account.
func (api *ManagementAPI) DeleteStorageAccount(storageAccountName string) error {
    _, err := api.session.delete("services/storageservices/" + storageAccountName)
    return err
}

// GetStorageAccountKeys retrieves a storage account's primary and secondary
// access keys from the Azure service.
func (api *ManagementAPI) GetStorageAccountKeys(accountName string) (*StorageAccountKeys, error) {
    url := "services/storageservices/" + accountName + "/keys"
    res, err := api.session.get(url)
    if err != nil {
        return nil, err
    }
    keys := StorageAccountKeys{}
    err = keys.Deserialize(res.Body)
    if err != nil {
        return nil, err
    }
    return &keys, nil
}

func (api *ManagementAPI) DeleteDisk(diskName string) error {
    _, err := api.session.delete("services/disks/" + diskName)
    return err
}

// Perform an operation on the specified role (as defined by serviceName, deploymentName and
// roleName)
// This is an asynchronous operation, it blocks until the operation is completed.
func (api *ManagementAPI) PerformRoleOperation(serviceName string, deploymentName string, roleName string, operation *RoleOperation) error {
    URI := "services/hostedservices/" + serviceName + "/deployments/" + deploymentName + "/roleinstances/" + roleName + "/Operations"
    body, err := marshalXML(operation)
    if err != nil {
        return err
    }
    response, err := api.session.post(URI, []byte(body), "application/xml")
    if err != nil {
        return err
    }
    return api.blockUntilCompleted(response)
}
