// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "encoding/xml"
    "fmt"
    . "launchpad.net/gocheck"
    "net/http"
    "time"
)

type managementAPIUtilitiesSuite struct {
    oldPollerInterval time.Duration
}

var _ = Suite(&managementAPIUtilitiesSuite{})

func (suite *managementAPIUtilitiesSuite) SetUpTest(c *C) {
    // Record the original poller interval. Will be restored at the end of
    // each test.
    suite.oldPollerInterval = pollerInterval
    // Set the poller interval to something short to speed up the tests.
    pollerInterval = time.Millisecond
}

func (suite *managementAPIUtilitiesSuite) TearDownTest(c *C) {
    // Restore old poller interval.
    pollerInterval = suite.oldPollerInterval
}

func makeX509ResponseWithOperationHeader(operationID string) *x509Response {
    header := http.Header{}
    header.Set(operationIDHeaderName, operationID)
    response := x509Response{
        StatusCode: http.StatusAccepted,
        Header:     header,
    }
    return &response
}

func makeAPI(c *C) *ManagementAPI {
    subscriptionId := "subscriptionId"
    api, err := NewManagementAPI(subscriptionId, "certFile")
    c.Assert(err, IsNil)
    return api
}

func (suite *managementAPIUtilitiesSuite) TestGetOperationIDExtractsHeader(c *C) {
    operationID := "operationID"
    response := makeX509ResponseWithOperationHeader(operationID)

    returnedOperationID, err := getOperationID(response)
    c.Assert(err, IsNil)
    c.Check(returnedOperationID, Equals, operationID)
}

func (suite *managementAPIUtilitiesSuite) TestBlockUntilCompletedErrorsIfOperationFails(c *C) {
    response := DispatcherResponse{
        response: &x509Response{
            Body:       []byte(fmt.Sprintf(operationXMLTemplate, "Failed")),
            StatusCode: http.StatusOK,
        },
        errorObject: nil}
    responses := []DispatcherResponse{response}
    rigPreparedResponseDispatcher(responses)
    operationID := "operationID"
    operationResponse := makeX509ResponseWithOperationHeader(operationID)
    api := makeAPI(c)

    err := api.blockUntilCompleted(operationResponse)
    c.Check(err, ErrorMatches, ".*asychronous operation failed.*")
}

func (suite *managementAPIUtilitiesSuite) TestBlockUntilCompletedErrorsIfInvalidXML(c *C) {
    response := DispatcherResponse{
        response: &x509Response{
            Body:       []byte(">invalidXML<"),
            StatusCode: http.StatusOK,
        },
        errorObject: nil}
    responses := []DispatcherResponse{response}
    rigPreparedResponseDispatcher(responses)
    operationID := "operationID"
    operationResponse := makeX509ResponseWithOperationHeader(operationID)
    api := makeAPI(c)

    err := api.blockUntilCompleted(operationResponse)
    c.Check(err, FitsTypeOf, new(xml.SyntaxError))
}

func (suite *managementAPIUtilitiesSuite) TestBlockUntilCompletedErrorsIfFailsToPoll(c *C) {
    response := DispatcherResponse{
        response:    &x509Response{},
        errorObject: fmt.Errorf("unexpected error")}
    responses := []DispatcherResponse{response}
    rigPreparedResponseDispatcher(responses)
    operationID := "operationID"
    operationResponse := makeX509ResponseWithOperationHeader(operationID)
    api := makeAPI(c)

    err := api.blockUntilCompleted(operationResponse)
    c.Check(err, ErrorMatches, ".*unexpected error.*")
}

func (suite *managementAPIUtilitiesSuite) TestBlockUntilCompletedSucceedsIfOperationSucceeds(c *C) {
    response := DispatcherResponse{
        response: &x509Response{
            Body:       []byte(fmt.Sprintf(operationXMLTemplate, "Succeeded")),
            StatusCode: http.StatusOK,
        },
        errorObject: nil}
    responses := []DispatcherResponse{response}
    rigPreparedResponseDispatcher(responses)
    operationID := "operationID"
    operationResponse := makeX509ResponseWithOperationHeader(operationID)
    api := makeAPI(c)

    err := api.blockUntilCompleted(operationResponse)
    c.Assert(err, IsNil)
}

type managementAPISuite struct {
    x509DispatcherFixture
    oldPollerInterval time.Duration
}

var _ = Suite(&managementAPISuite{})

func (suite *managementAPISuite) SetUpTest(c *C) {
    // Record the original poller interval. Will be restored at the end of
    // each test.
    suite.oldPollerInterval = pollerInterval
    // Set the poller interval to 0 to disable the polling.
    pollerInterval = 0
}

func (suite *managementAPISuite) TearDownTest(c *C) {
    // Restore old poller interval.
    pollerInterval = suite.oldPollerInterval
}

func (suite *managementAPISuite) TestNewManagementAPI(c *C) {
    subscriptionId := "subscriptionId"
    certFile := "certFile"

    api, err := NewManagementAPI(subscriptionId, certFile)

    c.Assert(err, IsNil)
    session, err := newX509Session(subscriptionId, certFile)
    c.Assert(api.session.subscriptionId, DeepEquals, session.subscriptionId)
    c.Assert(api.session.certFile, DeepEquals, session.certFile)
}

// setUpDispatcher sets up a request dispatcher that:
// - records requests
// - returns empty responses
func (suite *managementAPISuite) setUpDispatcher() *[]*x509Request {
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte{},
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*x509Request, 0)
    rigRecordingDispatcher(&recordedRequests)
    return &recordedRequests
}

// checkOneRequest asserts that the given slice contains one request, with the
// given characteristics.
func checkOneRequest(c *C, recordedRequests *[]*x509Request, URL string, payload []byte, Method string) {
    requests := *recordedRequests
    c.Assert(len(requests), Equals, 1)
    request := requests[0]
    checkRequest(c, request, URL, payload, Method)
}

func checkRequest(c *C, request *x509Request, URL string, payload []byte, Method string) {
    c.Assert(request.URL, Equals, URL)
    c.Assert(string(request.Payload), Equals, string(payload))
    c.Assert(request.Method, Equals, Method)
}

func (suite *managementAPISuite) TestGetOperationFailsIfNoHeader(c *C) {
    response := x509Response{
        StatusCode: http.StatusOK,
    }

    _, err := getOperationID(&response)
    c.Check(err, ErrorMatches, ".*no operation header.*")
}

func (suite *managementAPISuite) TestListHostedServiceDescriptors(c *C) {
    api := makeAPI(c)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte("<HostedServices></HostedServices>"),
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*x509Request, 0)
    rigRecordingDispatcher(&recordedRequests)

    _, err := api.ListHostedServiceDescriptors()

    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices"
    checkOneRequest(c, &recordedRequests, expectedURL, []byte{}, "GET")
}

// TODO test that ListHostedServiceDescriptors parses the structures correctly

func (suite *managementAPISuite) TestLoadHostedService(c *C) {
    api := makeAPI(c)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte("<HostedService></HostedService>"),
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*x509Request, 0)
    rigRecordingDispatcher(&recordedRequests)

    serviceName := "serviceName"

    _, err := api.LoadHostedService(serviceName)
    c.Assert(err, IsNil)

    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices/" + serviceName + "?embed-detail=true"
    checkOneRequest(c, &recordedRequests, expectedURL, []byte{}, "GET")
}

// TODO test that TestLoadHostedService parses the structures correctly

func (suite *managementAPISuite) TestAddHostedService(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    createHostedService := NewCreateHostedServiceWithLocation("testName", "testLabel", "East US")
    err := api.AddHostedService(createHostedService)
    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices"
    expectedPayload, err := marshalXML(createHostedService)
    c.Assert(err, IsNil)
    checkOneRequest(c, recordedRequests, expectedURL, expectedPayload, "POST")
}

func (suite *managementAPISuite) TestDeleteHostedService(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    hostedServiceName := "testName"
    err := api.DeleteHostedService(hostedServiceName)

    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices/" + hostedServiceName
    checkOneRequest(c, recordedRequests, expectedURL, []byte{}, "DELETE")
}

func (suite *managementAPISuite) TestAddDeployment(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    serviceName := "serviceName"
    configurationSet := NewLinuxProvisioningConfiguration("testHostname12345", "test", "test123#@!", false)
    OSVirtualHardDisk := NewOSVirtualHardDisk("hostCaching", "diskLabel", "diskName", "http://mediaLink", "sourceImageName")
    role := NewRole("ExtraSmall", "test-role-123", []LinuxProvisioningConfiguration{*configurationSet}, OSVirtualHardDisk)
    deployment := NewDeploymentForCreateVMDeployment("test-machine-name", "Staging", "testLabel", []Role{*role}, "testNetwork")
    err := api.AddDeployment(deployment, serviceName)

    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices/" + serviceName + "/deployments"
    expectedPayload, err := marshalXML(deployment)
    c.Assert(err, IsNil)
    checkOneRequest(c, recordedRequests, expectedURL, expectedPayload, "POST")
}

func (suite *managementAPISuite) TestDeleteDeployment(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    hostedServiceName := "testHosterServiceName"
    deploymentName := "testDeploymentName"
    err := api.DeleteDeployment(hostedServiceName, deploymentName)

    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices/" + hostedServiceName + "/deployments/" + deploymentName
    checkOneRequest(c, recordedRequests, expectedURL, []byte{}, "DELETE")
}

func (suite *managementAPISuite) TestAddStorageAccount(c *C) {
    api := makeAPI(c)
    header := http.Header{}
    header.Set("X-Ms-Request-Id", "operationID")
    fixedResponse := x509Response{
        StatusCode: http.StatusAccepted,
        Header:     header,
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*x509Request, 0)
    rigRecordingDispatcher(&recordedRequests)
    cssi := NewCreateStorageServiceInputWithLocation("name", "label", "East US")

    err := api.AddStorageAccount(cssi)
    c.Assert(err, IsNil)

    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/storageservices"
    expectedPayload, err := marshalXML(cssi)
    c.Assert(err, IsNil)
    checkOneRequest(c, &recordedRequests, expectedURL, expectedPayload, "POST")
}

func (suite *managementAPISuite) TestDeleteStorageAccount(c *C) {
    const accountName = "myaccount"
    api := makeAPI(c)
    accountURL := api.session.composeURL("services/storageservices/" + accountName)
    recordedRequests := setUpDispatcher("operationID")

    err := api.DeleteStorageAccount(accountName)
    c.Assert(err, IsNil)

    checkOneRequest(c, recordedRequests, accountURL, nil, "DELETE")
}

func (suite *managementAPISuite) TestGetStorageAccountKeys(c *C) {
    const (
        accountName  = "accountname"
        primaryKey   = "primarykey"
        secondaryKey = "secondarykey"
    )
    api := makeAPI(c)
    url := api.session.composeURL("services/storageservices/" + accountName)
    body := fmt.Sprintf(
        `<StorageService>
            <Url>%s</Url>
            <StorageServiceKeys>
                <Primary>%s</Primary>
                <Secondary>%s</Secondary>
            </StorageServiceKeys>
        </StorageService>`,
        url, primaryKey, secondaryKey)
    rigFixedResponseDispatcher(&x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte(body),
    })

    keys, err := api.GetStorageAccountKeys("account")
    c.Assert(err, IsNil)

    c.Check(keys.Primary, Equals, primaryKey)
    c.Check(keys.Secondary, Equals, secondaryKey)
    c.Check(keys.Url, Equals, url)
}

func (suite *managementAPISuite) TestDeleteDisk(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    diskName := "diskName"
    err := api.DeleteDisk(diskName)

    c.Assert(err, IsNil)
    expectedURL := api.session.composeURL("services/disks/" + diskName)
    checkOneRequest(c, recordedRequests, expectedURL, []byte{}, "DELETE")
}

func (suite *managementAPISuite) TestPerformNodeOperation(c *C) {
    api := makeAPI(c)
    recordedRequests := setUpDispatcher("operationID")
    serviceName := "serviceName"
    deploymentName := "deploymentName"
    roleName := "roleName"
    operation := newRoleOperation("RandomOperation")
    err := api.PerformRoleOperation(serviceName, deploymentName, roleName, operation)

    c.Assert(err, IsNil)
    expectedURL := AZURE_URL + api.session.subscriptionId + "/services/hostedservices/" + serviceName + "/deployments/" + deploymentName + "/roleinstances/" + roleName + "/Operations"
    expectedPayload, err := marshalXML(operation)
    c.Assert(err, IsNil)
    checkOneRequest(c, recordedRequests, expectedURL, expectedPayload, "POST")
}
