// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "bytes"
    "encoding/base64"
    "fmt"
    "io/ioutil"
    . "launchpad.net/gocheck"
    "launchpad.net/gwacl/dedent"
    "net/http"
    "net/url"
    "strings"
    "time"
)

type testComposeHeaders struct{}

var _ = Suite(&testComposeHeaders{})

func (suite *testComposeHeaders) TestNoHeaders(c *C) {
    req, err := http.NewRequest("GET", "http://example.com", nil)
    c.Assert(err, IsNil)

    observed := composeHeaders(req)
    expected := "\n\n\n\n\n\n\n\n\n\n\n"

    c.Assert(observed, Equals, expected)
}

func (suite *testComposeHeaders) TestCreatesHeaders(c *C) {
    req, err := http.NewRequest("GET", "http://example.com", nil)
    c.Assert(err, IsNil)

    var items []string
    for i, header_name := range headers_to_sign {
        v := fmt.Sprintf("%d", i)
        req.Header.Set(header_name, v)
        items = append(items, v+"\n")
    }
    expected := strings.Join(items, "")

    observed := composeHeaders(req)
    c.Assert(observed, Equals, expected)
}

func (suite *testComposeHeaders) TestCanonicalizedHeaders(c *C) {
    req, err := http.NewRequest("GET", "http://example.com", nil)
    c.Assert(err, IsNil)
    req.Header.Set("x-ms-why", "aye")
    req.Header.Set("x-ms-foo", "bar")
    req.Header.Set("invalid", "blah")

    expected := "x-ms-foo:bar\nx-ms-why:aye\n"
    observed := composeCanonicalizedHeaders(req)
    c.Check(observed, Equals, expected)
}

type TestComposeCanonicalizedResource struct{}

var _ = Suite(&TestComposeCanonicalizedResource{})

func (suite *TestComposeCanonicalizedResource) TestPrependsSlash(c *C) {
    req, err := http.NewRequest("GET", "http://example.com", nil)
    c.Assert(err, IsNil)
    path := MakeRandomString(10)
    req.URL.Path = path
    account_name := MakeRandomString(10)
    observed := composeCanonicalizedResource(req, account_name)
    expected := "/" + account_name + "/" + path
    c.Assert(observed, Equals, expected)
}

func (suite *TestComposeCanonicalizedResource) TestCreatesResource(c *C) {
    path := MakeRandomString(5)
    req, err := http.NewRequest("GET", fmt.Sprintf("http://example.com/%s", path), nil)
    c.Assert(err, IsNil)

    account_name := MakeRandomString(10)
    observed := composeCanonicalizedResource(req, account_name)
    expected := "/" + account_name + "/" + path
    c.Assert(observed, Equals, expected)
}

func (suite *TestComposeCanonicalizedResource) TestQueryParams(c *C) {
    req, err := http.NewRequest(
        "GET", "http://example.com/?Kevin=Perry&foo=bar", nil)
    c.Assert(err, IsNil)

    account_name := MakeRandomString(10)
    observed := composeCanonicalizedResource(req, account_name)
    expected := "/" + account_name + "/\n" + "foo:bar" + "\n" + "kevin:Perry"
    c.Assert(observed, Equals, expected)
}

func (suite *TestComposeCanonicalizedResource) TestToLowerKeys(c *C) {
    values := url.Values{
        "foo":   []string{"bar", "baz"},
        "alpha": []string{"gamma", "delta"},
        "quux":  []string{"flobble"},
        "BIG":   []string{"Big", "Data"},
        "big":   []string{"Big", "Little"},
    }

    observed := toLowerKeys(values)
    expected := map[string][]string{
        "foo":   {"bar", "baz"},
        "alpha": {"delta", "gamma"},
        "quux":  {"flobble"},
        "big":   {"Big", "Big", "Data", "Little"},
    }

    c.Check(observed, DeepEquals, expected)
}

func (suite *TestComposeCanonicalizedResource) TestEncodeParams(c *C) {
    input := map[string][]string{
        "foo":   {"bar", "baz"},
        "alpha": {"delta", "gamma"},
        "quux":  {"flobble"},
        "big":   {"Big", "Big", "Data", "Little"},
    }

    observed := encodeParams(input)
    expected := ("alpha:delta,gamma\nbig:Big,Big,Data,Little\n" +
        "foo:bar,baz\nquux:flobble")
    c.Assert(observed, Equals, expected)
}

func (suite *TestComposeCanonicalizedResource) TestEncodeParamsEmpty(c *C) {
    input := map[string][]string{}
    observed := encodeParams(input)
    expected := ""
    c.Assert(observed, Equals, expected)
}

type TestComposeStringToSign struct{}

var _ = Suite(&TestComposeStringToSign{})

func (suite *TestComposeStringToSign) TestFullRequest(c *C) {
    req, err := http.NewRequest(
        "GET", "http://example.com/mypath?Kevin=Perry&foo=bar", nil)
    c.Assert(err, IsNil)
    for i, header_name := range headers_to_sign {
        req.Header.Set(header_name, fmt.Sprintf("%v", i))
    }
    req.Header.Set("x-ms-testing", "foo")
    expected := "GET\n0\n1\n2\n3\n4\n5\n6\n7\n8\n9\n10\nx-ms-testing:foo\n/myaccount/mypath\nfoo:bar\nkevin:Perry"
    observed := composeStringToSign(req, "myaccount")
    c.Assert(observed, Equals, expected)
}

type TestComposeAuthHeader struct{}

var _ = Suite(&TestComposeAuthHeader{})

func (suite *TestComposeAuthHeader) TestCreatesHeaderString(c *C) {
    req, err := http.NewRequest(
        "GET", "http://example.com/mypath?Kevin=Perry&foo=bar", nil)
    c.Assert(err, IsNil)

    key := base64.StdEncoding.EncodeToString([]byte("dummykey"))

    observed := composeAuthHeader(req, "myname", key)
    expected := "SharedKey myname:Xf9hWQ99mM0IyEOL6rNeAUdTQlixVqiYnt2TpLCCpY0="
    c.Assert(observed, Equals, expected)
}

type TestSignRequest struct{}

var _ = Suite(&TestSignRequest{})

func (suite *TestSignRequest) TestAddsHeaderToRequest(c *C) {
    req, err := http.NewRequest(
        "GET", "http://example.com/mypath?Kevin=Perry&foo=bar", nil)
    c.Assert(err, IsNil)
    c.Assert(req.Header.Get("Authorization"), Equals, "")

    key := base64.StdEncoding.EncodeToString([]byte("dummykey"))
    signRequest(req, "myname", key)

    expected := "SharedKey myname:Xf9hWQ99mM0IyEOL6rNeAUdTQlixVqiYnt2TpLCCpY0="
    c.Assert(req.Header.Get("Authorization"), Equals, expected)
}

type TestRequestHeaders struct{}

var _ = Suite(&TestRequestHeaders{})

func (suite *TestRequestHeaders) TestAddsVersionHeaderToRequest(c *C) {
    req, err := http.NewRequest("GET", "http://example.com/", nil)
    c.Assert(err, IsNil)
    addVersionHeader(req, "2012-02-12")
    c.Assert(req.Header.Get("x-ms-version"), Equals, "2012-02-12")
}

func (suite *TestRequestHeaders) TestContentHeader(c *C) {
    data := "test data"
    req, err := http.NewRequest("PUT", "http://example.com/", strings.NewReader(data))
    c.Assert(err, IsNil)
    addContentHeaders(req)
    c.Assert(
        req.Header.Get("Content-Length"), Equals, fmt.Sprintf("%d", len(data)))

    // Ensure that reading the request data didn't destroy it.
    reqdata, _ := ioutil.ReadAll(req.Body)
    c.Assert(data, Equals, string(reqdata))
}

func (suite *TestRequestHeaders) TestLengthHeaderNotSetForGET(c *C) {
    req, err := http.NewRequest("GET", "http://example.com/", nil)
    c.Assert(err, IsNil)
    addContentHeaders(req)
    _, length_present := req.Header[http.CanonicalHeaderKey("Content-Length")]
    c.Assert(length_present, Equals, false)
}

func (suite *TestRequestHeaders) TestContentHeaderWithNoBody(c *C) {
    req, err := http.NewRequest("PUT", "http://example.com/", nil)
    c.Assert(err, IsNil)
    addContentHeaders(req)
    _, md5_present := req.Header[http.CanonicalHeaderKey("Content-MD5")]
    c.Check(md5_present, Equals, false)
    content := req.Header.Get("Content-Length")
    c.Check(content, Equals, "0")
}

func (suite *TestRequestHeaders) TestDateHeader(c *C) {
    req, err := http.NewRequest("GET", "http://example.com/", nil)
    c.Assert(err, IsNil)
    c.Assert(req.Header.Get("Date"), Equals, "")
    addDateHeader(req)
    observed := req.Header.Get("Date")
    observed_time, err := time.Parse(time.RFC1123, observed)
    c.Assert(err, IsNil)
    difference := time.Now().UTC().Sub(observed_time)
    if difference.Minutes() > 1.0 {
        c.FailNow()
    }
}

type TestStorageContext struct{}

var _ = Suite(&TestStorageContext{})

func (suite *TestStorageContext) TestGetClientReturnsDefaultClient(c *C) {
    context := &StorageContext{client: nil}
    c.Assert(context.getClient(), Equals, http.DefaultClient)
}

func (suite *TestStorageContext) TestGetClientReturnsSpecifiedClient(c *C) {
    context := &StorageContext{client: &http.Client{}}
    c.Assert(context.getClient(), Not(Equals), http.DefaultClient)
    c.Assert(context.getClient(), Equals, context.client)
}

// TestTransport is used as an http.Client.Transport for testing. The only
// requirement is that it adhere to the http.RoundTripper interface.
type TestTransport struct {
    Request  *http.Request
    Response *http.Response
    Error    error
}

func (t *TestTransport) RoundTrip(req *http.Request) (resp *http.Response, err error) {
    t.Request = req
    return t.Response, t.Error
}

// Convenience factory to create a StorageContext with a random name and
// random base64-encoded key.
func makeStorageContext() *StorageContext {
    return &StorageContext{
        Account: MakeRandomString(10),
        Key:     base64.StdEncoding.EncodeToString(MakeRandomByteSlice(10)),
    }
}

type TestListContainers struct{}

var _ = Suite(&TestListContainers{})

// The ListContainers Storage API call returns a ContainerEnumerationResults
// struct on success.
func (suite *TestListContainers) Test(c *C) {
    response_body := `
        <?xml version="1.0" encoding="utf-8"?>
        <EnumerationResults AccountName="http://myaccount.blob.core.windows.net">
          <Prefix>prefix-value</Prefix>
          <Marker>marker-value</Marker>
          <MaxResults>max-results-value</MaxResults>
          <Containers>
            <Container>
              <Name>name-value</Name>
              <URL>url-value</URL>
              <Properties>
                <Last-Modified>date/time-value</Last-Modified>
                <Etag>etag-value</Etag>
                <LeaseStatus>lease-status-value</LeaseStatus>
                <LeaseState>lease-state-value</LeaseState>
                <LeaseDuration>lease-duration-value</LeaseDuration>
              </Properties>
              <Metadata>
                <metadata-name>metadata-value</metadata-name>
              </Metadata>
            </Container>
          </Containers>
          <NextMarker>next-marker-value</NextMarker>
        </EnumerationResults>`
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusOK),
        StatusCode: http.StatusOK,
        Body:       ioutil.NopCloser(strings.NewReader(response_body)),
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    results, err := context.ListContainers()
    c.Assert(err, IsNil)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/?comp=list", context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
    c.Assert(results, NotNil)
    c.Assert(results.Containers[0].Name, Equals, "name-value")
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestListContainers) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.ListContainers()
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestListContainers) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusNotFound),
        StatusCode: http.StatusNotFound,
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.ListContainers()
    c.Assert(err, NotNil)
}

type TestListBlobs struct{}

var _ = Suite(&TestListBlobs{})

// The ListBlobs Storage API call returns a BlobEnumerationResults struct on
// success.
func (suite *TestListBlobs) Test(c *C) {
    response_body := `
        <?xml version="1.0" encoding="utf-8"?>
        <EnumerationResults ContainerName="http://myaccount.blob.core.windows.net/mycontainer">
          <Prefix>prefix</Prefix>
          <Marker>marker</Marker>
          <MaxResults>maxresults</MaxResults>
          <Delimiter>delimiter</Delimiter>
          <Blobs>
            <Blob>
              <Name>blob-name</Name>
              <Snapshot>snapshot-date-time</Snapshot>
              <Url>blob-address</Url>
              <Properties>
                <Last-Modified>last-modified</Last-Modified>
                <Etag>etag</Etag>
                <Content-Length>size-in-bytes</Content-Length>
                <Content-Type>blob-content-type</Content-Type>
                <Content-Encoding />
                <Content-Language />
                <Content-MD5 />
                <Cache-Control />
                <x-ms-blob-sequence-number>sequence-number</x-ms-blob-sequence-number>
                <BlobType>blobtype</BlobType>
                <LeaseStatus>leasestatus</LeaseStatus>
                <LeaseState>leasestate</LeaseState>
                <LeaseDuration>leasesduration</LeaseDuration>
                <CopyId>id</CopyId>
                <CopyStatus>copystatus</CopyStatus>
                <CopySource>copysource</CopySource>
                <CopyProgress>copyprogress</CopyProgress>
                <CopyCompletionTime>copycompletiontime</CopyCompletionTime>
                <CopyStatusDescription>copydesc</CopyStatusDescription>
              </Properties>
              <Metadata>
                <MetaName1>metadataname1</MetaName1>
                <MetaName2>metadataname2</MetaName2>
              </Metadata>
            </Blob>
            <BlobPrefix>
              <Name>blob-prefix</Name>
            </BlobPrefix>
          </Blobs>
          <NextMarker />
        </EnumerationResults>`
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusOK),
        StatusCode: http.StatusOK,
        Body:       ioutil.NopCloser(strings.NewReader(response_body)),
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    results, err := context.ListBlobs("container")
    c.Assert(err, IsNil)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container?restype=container&comp=list",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
    c.Assert(results, NotNil)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestListBlobs) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.ListBlobs("container")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestListBlobs) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusNotFound),
        StatusCode: http.StatusNotFound,
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.ListBlobs("container")
    c.Assert(err, NotNil)
}

type TestCreateContainer struct{}

var _ = Suite(&TestCreateContainer{})

// The CreateContainer Storage API call returns without error when the
// container has been created successfully.
func (suite *TestCreateContainer) Test(c *C) {
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusCreated),
        StatusCode: http.StatusCreated,
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    container_name := MakeRandomString(10)
    err := context.CreateContainer(container_name)
    c.Assert(err, IsNil)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/%s?restype=container",
        context.Account, container_name))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestCreateContainer) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.CreateContainer("container")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestCreateContainer) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusNotFound),
        StatusCode: http.StatusNotFound,
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.CreateContainer("container")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestCreateContainer) TestNotCreatedResponse(c *C) {
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusOK),
        StatusCode: http.StatusOK,
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.CreateContainer("container")
    c.Assert(err, NotNil)
}

type TestPutBlob struct{}

var _ = Suite(&TestPutBlob{})

// The ListBlobs Storage API call returns 201 CREATED on success.
func (suite *TestPutBlob) Test(c *C) {
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusCreated),
        StatusCode: http.StatusCreated,
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    err := context.PutBlob("container", "blobname")
    c.Assert(err, IsNil)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/blobname",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestPutBlob) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.PutBlob("container", "blobname")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestPutBlob) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     "102 Frotzed",
        StatusCode: 102,
        Body:       ioutil.NopCloser(strings.NewReader("<Error><Code>Frotzed</Code><Message>failed to put blob</Message></Error>")),
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.PutBlob("container", "blobname")
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, ".*102.*")
    c.Check(err, ErrorMatches, ".*Frotzed.*")
    c.Check(err, ErrorMatches, ".*failed to put blob.*")
}

type TestPutBlock struct{}

var _ = Suite(&TestPutBlock{})

func (suite *TestPutBlock) Test(c *C) {
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusCreated),
        StatusCode: http.StatusCreated,
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    blockid := "\x1b\xea\xf7Mv\xb5\xddH\xebm"
    random_data := MakeRandomByteSlice(10)
    data_reader := bytes.NewReader(random_data)
    err := context.PutBlock("container", "blobname", blockid, data_reader)
    c.Assert(err, IsNil)

    // The blockid should have been base64 encoded and url escaped.
    base64_id := base64.StdEncoding.EncodeToString([]byte(blockid))
    escaped_id := url.QueryEscape(base64_id)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/blobname?comp=block&blockid=%s",
        context.Account, escaped_id))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")

    data, err := ioutil.ReadAll(transport.Request.Body)
    c.Assert(err, IsNil)
    c.Check(data, DeepEquals, random_data)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestPutBlock) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    data_reader := bytes.NewReader(MakeRandomByteSlice(10))
    err := context.PutBlock("container", "blobname", "blockid", data_reader)
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestPutBlock) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     "102 Frotzed",
        StatusCode: 102,
        Body:       ioutil.NopCloser(strings.NewReader("<Error><Code>Frotzed</Code><Message>failed to put block</Message></Error>")),
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    data_reader := bytes.NewReader(MakeRandomByteSlice(10))
    err := context.PutBlock("container", "blobname", "blockid", data_reader)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, ".*102.*")
    c.Check(err, ErrorMatches, ".*Frotzed.*")
    c.Check(err, ErrorMatches, ".*failed to put block.*")
}

type TestPutBlockList struct{}

var _ = Suite(&TestPutBlockList{})

func (suite *TestPutBlockList) Test(c *C) {
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusCreated),
        StatusCode: http.StatusCreated,
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    blocklist := &BlockList{}
    blocklist.Add(BlockListLatest, "b1")
    blocklist.Add(BlockListLatest, "b2")
    err := context.PutBlockList("container", "blobname", blocklist)
    c.Assert(err, IsNil)

    c.Check(transport.Request.Method, Equals, "PUT")
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/blobname?comp=blocklist",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")

    data, err := ioutil.ReadAll(transport.Request.Body)
    c.Assert(err, IsNil)
    expected := dedent.Dedent(`
        <BlockList>
          <Latest>YjE=</Latest>
          <Latest>YjI=</Latest>
        </BlockList>`)
    c.Check(string(data), Equals, expected)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestPutBlockList) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    blocklist := &BlockList{}
    err := context.PutBlockList("container", "blobname", blocklist)
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestPutBlockList) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     "102 Frotzed",
        StatusCode: 102,
        Body:       ioutil.NopCloser(strings.NewReader("<Error><Code>Frotzed</Code><Message>failed to put blocklist</Message></Error>")),
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    blocklist := &BlockList{}
    err := context.PutBlockList("container", "blobname", blocklist)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, ".*102.*")
    c.Check(err, ErrorMatches, ".*Frotzed.*")
    c.Check(err, ErrorMatches, ".*failed to put blocklist.*")
}

type TestGetBlockList struct{}

var _ = Suite(&TestGetBlockList{})

// The GetBlockList Storage API call returns a GetBlockList struct on
// success.
func (suite *TestGetBlockList) Test(c *C) {
    response_body := `
        <?xml version="1.0" encoding="utf-8"?>
        <BlockList>
          <CommittedBlocks>
            <Block>
              <Name>BlockId001</Name>
              <Size>4194304</Size>
            </Block>
          </CommittedBlocks>
          <UncommittedBlocks>
            <Block>
              <Name>BlockId002</Name>
              <Size>1024</Size>
            </Block>
          </UncommittedBlocks>
        </BlockList>`

    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusOK),
        StatusCode: http.StatusOK,
        Body:       ioutil.NopCloser(strings.NewReader(response_body)),
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    results, err := context.GetBlockList("container", "myfilename")
    c.Assert(err, IsNil)
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/myfilename?comp=blocklist&blocklisttype=all",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
    c.Assert(results, NotNil)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestGetBlockList) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.GetBlockList("container", "myfilename")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestGetBlockList) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusNotFound),
        StatusCode: http.StatusNotFound,
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    _, err := context.GetBlockList("container", "myfilename")
    c.Assert(err, NotNil)
}

type TestDeleteBlob struct{}

var _ = Suite(&TestDeleteBlob{})

func (suite *TestDeleteBlob) Test(c *C) {
    context := makeStorageContext()
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusAccepted),
        StatusCode: http.StatusAccepted,
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    err := context.DeleteBlob("container", "blobname")
    c.Assert(err, IsNil)

    c.Check(transport.Request.Method, Equals, "DELETE")
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/blobname",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")
    c.Check(transport.Request.Body, IsNil)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestDeleteBlob) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.DeleteBlob("container", "blobname")
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestDeleteBlob) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     "146 Frotzed",
        StatusCode: 146,
        Body:       ioutil.NopCloser(strings.NewReader("<Error><Code>Frotzed</Code><Message>failed to delete blob</Message></Error>")),
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    err := context.DeleteBlob("container", "blobname")
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, ".*146.*")
    c.Check(err, ErrorMatches, ".*Frotzed.*")
    c.Check(err, ErrorMatches, ".*failed to delete blob.*")
}

type TestGetBlob struct{}

var _ = Suite(&TestGetBlob{})

func (suite *TestGetBlob) Test(c *C) {
    context := makeStorageContext()
    response_body := "blob-in-a-can"
    response := &http.Response{
        Status:     fmt.Sprintf("%d", http.StatusOK),
        StatusCode: http.StatusOK,
        Body:       ioutil.NopCloser(strings.NewReader(response_body)),
    }
    transport := &TestTransport{Response: response}
    context.client = &http.Client{Transport: transport}
    reader, err := context.GetBlob("container", "blobname")
    c.Assert(err, IsNil)
    c.Assert(reader, NotNil)
    defer reader.Close()

    c.Check(transport.Request.Method, Equals, "GET")
    c.Check(transport.Request.URL.String(), Equals, fmt.Sprintf(
        "http://%s.blob.core.windows.net/container/blobname",
        context.Account))
    c.Check(transport.Request.Header.Get("Authorization"), Not(Equals), "")

    data, err := ioutil.ReadAll(reader)
    c.Assert(err, IsNil)
    c.Check(string(data), Equals, response_body)
}

// Client-side errors from the HTTP client are propagated back to the caller.
func (suite *TestGetBlob) TestError(c *C) {
    error := fmt.Errorf("canned-error")
    transport := &TestTransport{Error: error}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    reader, err := context.GetBlob("container", "blobname")
    c.Check(reader, IsNil)
    c.Assert(err, NotNil)
}

// Server-side errors are propagated back to the caller.
func (suite *TestGetBlob) TestErrorResponse(c *C) {
    response := &http.Response{
        Status:     "246 Frotzed",
        StatusCode: 246,
        Body:       ioutil.NopCloser(strings.NewReader("<Error><Code>Frotzed</Code><Message>failed to get blob</Message></Error>")),
    }
    transport := &TestTransport{Response: response}
    context := makeStorageContext()
    context.client = &http.Client{Transport: transport}
    reader, err := context.GetBlob("container", "blobname")
    c.Check(reader, IsNil)
    c.Assert(err, NotNil)
    c.Check(err, ErrorMatches, ".*246.*")
    c.Check(err, ErrorMatches, ".*Frotzed.*")
    c.Check(err, ErrorMatches, ".*failed to get blob.*")
}
