// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "fmt"
    "math/rand"
    "net/http"
    "strings"
    "time"
)

// This should be refactored at some point, it does not belong in here.
// Perhaps we can add it to gocheck, or start a testtools-like project.
const chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890abcdefghijklmnopqrstuvwxyz"

func MakeRandomString(length int) string {
    return string(MakeRandomByteSlice(length))
}

func MakeRandomByteSlice(length int) []byte {
    dest := make([]byte, length)
    for i := range dest {
        num := rand.Intn(len(chars))
        rand_char := chars[num]
        dest[i] = rand_char
    }
    return dest
}

func MakeRandomBool() bool {
    v := rand.Intn(2)
    if v == 0 {
        return false
    }
    return true
}

func BoolToString(v bool) string {
    if v {
        return "true"
    }
    return "false"
}

func StringToBool(v string) bool {
    switch strings.ToLower(v) {
    case "true":
        return true
    case "false":
        return false
    }
    panic(fmt.Errorf("can't convert '%s' to a bool", v))
}

type endpointParams struct {
    setname   string
    localport int
    name      string
    port      int
    protocol  string
}

func makeEndpoint(params endpointParams) *InputEndpoint {
    if params.setname == "" {
        params.setname = MakeRandomString(10)
    }
    if params.name == "" {
        params.name = MakeRandomString(10)
    }
    if params.protocol == "" {
        params.protocol = MakeRandomString(10)
    }
    if params.localport == 0 {
        params.localport = rand.Intn(65535)
    }
    if params.port == 0 {
        params.port = rand.Intn(65535)
    }

    return &InputEndpoint{
        LoadBalancedEndpointSetName: params.setname,
        LocalPort:                   params.localport,
        Name:                        params.name,
        Port:                        params.port,
        Protocol:                    params.protocol}
}

func makeLinuxProvisioningConfiguration() *LinuxProvisioningConfiguration {
    hostname := MakeRandomString(10)
    username := MakeRandomString(10)
    password := MakeRandomString(10)
    disable_ssh := MakeRandomBool()

    return &LinuxProvisioningConfiguration{
        ConfigurationSetType: "LinuxProvisioningConfiguration",
        Hostname:             hostname,
        Username:             username,
        Password:             password,
        DisableSshPasswordAuthentication: disable_ssh}
}

func makeOSVirtualHardDisk() *OSVirtualHardDisk {
    HostCaching := BoolToString(MakeRandomBool())
    DiskLabel := MakeRandomString(10)
    DiskName := MakeRandomString(10)
    MediaLink := MakeRandomString(10)
    SourceImageName := MakeRandomString(10)

    return &OSVirtualHardDisk{
        HostCaching:     HostCaching,
        DiskLabel:       DiskLabel,
        DiskName:        DiskName,
        MediaLink:       MediaLink,
        SourceImageName: SourceImageName}
}

func makeRole() *Role {
    RoleSize := "ExtraSmall"
    RoleName := MakeRandomString(10)
    RoleType := "PersistentVMRole"
    config := makeLinuxProvisioningConfiguration()
    configset := []LinuxProvisioningConfiguration{*config}

    return &Role{
        RoleSize:          RoleSize,
        RoleName:          RoleName,
        RoleType:          RoleType,
        ConfigurationSets: configset}
}

func makeDnsServer() *DnsServer {
    name := MakeRandomString(10)
    address := MakeRandomString(10)

    return &DnsServer{
        Name:    name,
        Address: address}
}

func makeDeployment() *Deployment {
    Name := MakeRandomString(10)
    DeploymentSlot := "Staging"
    Label := MakeRandomString(10)
    VirtualNetworkName := MakeRandomString(10)
    role := makeRole()
    RoleList := []Role{*role}
    Dns := []DnsServer{*makeDnsServer()}

    return &Deployment{
        XMLNS:              XMLNS,
        XMLNS_I:            XMLNS_I,
        Name:               Name,
        DeploymentSlot:     DeploymentSlot,
        Label:              Label,
        RoleList:           RoleList,
        VirtualNetworkName: VirtualNetworkName,
        DNS:                Dns,
    }
}

func makeCreateStorageServiceInput() *CreateStorageServiceInput {
    ServiceName := MakeRandomString(10)
    Description := MakeRandomString(10)
    Label := MakeRandomString(10)
    AffinityGroup := MakeRandomString(10)
    Location := MakeRandomString(10)
    GeoReplicationEnabled := BoolToString(MakeRandomBool())
    ExtendedProperties := []ExtendedProperty{{
        Name:  MakeRandomString(10),
        Value: MakeRandomString(10)}}

    return &CreateStorageServiceInput{
        XMLNS:                 XMLNS,
        ServiceName:           ServiceName,
        Description:           Description,
        Label:                 Label,
        AffinityGroup:         AffinityGroup,
        Location:              Location,
        GeoReplicationEnabled: GeoReplicationEnabled,
        ExtendedProperties:    ExtendedProperties}
}

// rigRecordingDispatcher sets up a request dispatcher that records incoming
// requests by appending them to *record.  It returns the result of whatever
// dispatcher was already active.
// If you also want the dispatcher to return a particular result, rig it for
// that result first (using one of the other rig...Dispatcher functions) and
// then chain the recording dispatcher in front of it.
func rigRecordingDispatcher(record *[]*x509Request) {
    previousDispatcher := _X509Dispatcher
    _X509Dispatcher = func(session *x509Session, request *x509Request) (*x509Response, error) {
        *record = append(*record, request)
        return previousDispatcher(session, request)
    }
}

// rigFixedResponseDispatcher sets up a request dispatcher that always returns
// a prepared response.
func rigFixedResponseDispatcher(response *x509Response) {
    _X509Dispatcher = func(*x509Session, *x509Request) (*x509Response, error) {
        return response, nil
    }
}

// rigFailingDispatcher sets up a request dispatcher that returns a given
// error.
func rigFailingDispatcher(failure error) {
    _X509Dispatcher = func(*x509Session, *x509Request) (*x509Response, error) {
        return nil, failure
    }
}

type DispatcherResponse struct {
    response    *x509Response
    errorObject error
}

// rigPreparedResponseDispatcher sets up a request dispatcher that all the provided
// reponses, one after the other each time the dispatcher is called.
func rigPreparedResponseDispatcher(responses []DispatcherResponse) {
    index := 0
    _X509Dispatcher = func(*x509Session, *x509Request) (*x509Response, error) {
        response := responses[index]
        index += 1
        return response.response, response.errorObject
    }
}

// setUpDispatcher sets up a request dispatcher that:
// - records requests
// - returns empty responses
func setUpDispatcher(operationID string) *[]*x509Request {
    header := http.Header{}
    header.Set("X-Ms-Request-Id", operationID)
    fixedResponse := x509Response{
        StatusCode: http.StatusOK,
        Body:       []byte{},
        Header:     header,
    }
    rigFixedResponseDispatcher(&fixedResponse)
    recordedRequests := make([]*x509Request, 0)
    rigRecordingDispatcher(&recordedRequests)
    return &recordedRequests
}

func init() {
    // Staggeringly, rand will produce the same numbers from the start of
    // program invocation unless you seed it ...
    rand.Seed(int64(time.Now().Nanosecond()))
}
