// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import "encoding/xml"
import "encoding/base64"
import "fmt"

// It's impossible to have any kind of common method inherited into all the
// various serializable objects because the receiver of the method is the
// wrong class which confuses the xml marshaller.  Hence, this mess.
type AzureObject interface {
    Serialize() (string, error)
}

// marshalXML is a wrapper for serializing objects to XML in the visual layout
// that gwacl prefers.
func marshalXML(obj interface{}) ([]byte, error) {
    return xml.MarshalIndent(obj, "", "  ")
}

func toxml(obj AzureObject) (string, error) {
    out, err := marshalXML(obj)
    if err != nil {
        return "", err
    }
    return string(out), nil
}

//
// LinuxProvisioningConfiguration bits
//

// TODO: this should contain InputEndpoints objects.
type LinuxProvisioningConfiguration struct {
    ConfigurationSetType             string `xml:"ConfigurationSetType"` // "LinuxProvisioningConfiguration"
    Hostname                         string `xml:"HostName"`
    Username                         string `xml:"UserName"`
    Password                         string `xml:"UserPassword"`
    DisableSshPasswordAuthentication bool   `xml:"DisableSshPasswordAuthentication"`
}

func (c *LinuxProvisioningConfiguration) Serialize() (string, error) {
    return toxml(c)
}

func NewLinuxProvisioningConfiguration(
    Hostname, Username, Password string,
    DisableSshPasswordAuthentication bool) *LinuxProvisioningConfiguration {
    return &LinuxProvisioningConfiguration{
        ConfigurationSetType: "LinuxProvisioningConfiguration",
        Hostname:             Hostname,
        Username:             Username,
        Password:             Password,
        DisableSshPasswordAuthentication: DisableSshPasswordAuthentication,
    }
}

//
// InputEndpoint bits
//

type InputEndpoint struct {
    LoadBalancedEndpointSetName string `xml:"LoadBalancedEndpointSetName"`
    LocalPort                   int    `xml:"LocalPort"`
    Name                        string `xml:"Name"`
    Port                        int    `xml:"Port"`
    Protocol                    string `xml:"Protocol"` // TCP or UDP
}

func (c *InputEndpoint) Serialize() (string, error) {
    return toxml(c)
}

//
// OSVirtualHardDisk bits
//

type HostCachingType string

const (
    HostCachingRO HostCachingType = "ReadOnly"
    HostCachingRW HostCachingType = "ReadWrite"
)

type OSVirtualHardDisk struct {
    HostCaching     string `xml:"HostCaching,omitempty"`
    DiskLabel       string `xml:"DiskLabel,omitempty"`
    DiskName        string `xml:"DiskName,omitempty"`
    MediaLink       string `xml:"MediaLink,omitempty"`
    SourceImageName string `xml:"SourceImageName,omitempty"`
}

func (c *OSVirtualHardDisk) Serialize() (string, error) {
    return toxml(c)
}

func NewOSVirtualHardDisk(HostCaching HostCachingType, DiskLabel string,
    DiskName string, MediaLink string,
    SourceImageName string) *OSVirtualHardDisk {
    return &OSVirtualHardDisk{
        HostCaching:     string(HostCaching),
        DiskLabel:       DiskLabel,
        DiskName:        DiskName,
        SourceImageName: SourceImageName,
        MediaLink:       MediaLink,
    }
}

//
// NetworkConfiguration bits
//

type NetworkConfiguration struct {
    InputEndpoints []InputEndpoint `xml:"InputEndpoints>InputEndpoint"`
}

func (c *NetworkConfiguration) Serialize() (string, error) {
    return toxml(c)
}

type Role struct {
    RoleName          string                           `xml:"RoleName"`
    RoleType          string                           `xml:"RoleType"` // Always "PersistentVMRole"
    ConfigurationSets []LinuxProvisioningConfiguration `xml:"ConfigurationSets>ConfigurationSet"`
    OSVirtualHardDisk *OSVirtualHardDisk               `xml:"OSVirtualHardDisk"`
    RoleSize          string                           `xml:"RoleSize"`
}

//
// Role bits
//

func (c *Role) Serialize() (string, error) {
    return toxml(c)
}

func NewRole(RoleSize string, RoleName string,
    ConfigurationSets []LinuxProvisioningConfiguration, OSVirtualHardDisk *OSVirtualHardDisk) *Role {
    return &Role{
        RoleSize:          RoleSize,
        RoleName:          RoleName,
        RoleType:          "PersistentVMRole",
        ConfigurationSets: ConfigurationSets,
        OSVirtualHardDisk: OSVirtualHardDisk,
    }
}

//
// DnsServer bits
//

type DnsServer struct {
    Name    string `xml:"Name"`
    Address string `xml:"Address"`
}

func (c *DnsServer) Serialize() (string, error) {
    return toxml(c)
}

//
// Hosted service bits
//

type HostedService struct {
    XMLNS       string       `xml:"xmlns,attr"`
    Url         string       `xml:"Url"`
    ServiceName string       `xml:"ServiceName"`
    Description string       `xml:"Description"`
    Deployments []Deployment `xml:"Deployments>Deployment"`
}

type HostedServiceDescriptorList struct {
    HostedServices []HostedServiceDescriptor `xml:"HostedService"`
}

type HostedServiceDescriptor struct {
    XMLNS       string `xml:"xmlns,attr"`
    Url         string `xml:"Url"`
    ServiceName string `xml:"ServiceName"`
    Description string `xml:"Description"`
}

type CreateHostedService struct {
    // See http://msdn.microsoft.com/en-us/library/windowsazure/hh264518.aspx
    XMLNS       string `xml:"xmlns,attr"`
    ServiceName string `xml:"ServiceName"`
    // Label is base64-encoded.
    Label              string             `xml:"Label"`
    Description        string             `xml:"Description,omitempty"`
    AffinityGroup      string             `xml:"AffinityGroup,omitempty"`
    Location           string             `xml:"Location,omitempty"`
    ExtendedProperties []ExtendedProperty `xml:"ExtendedProperties>ExtendedProperty,omitempty"`
}

func NewCreateHostedServiceWithLocation(serviceName, label, location string) *CreateHostedService {
    base64label := base64.StdEncoding.EncodeToString([]byte(label))
    return &CreateHostedService{
        XMLNS:       XMLNS,
        ServiceName: serviceName,
        Label:       base64label,
        Location:    location,
    }
}

//
// Deployment bits
//

// Deployment is used both as input for the "Create Virtual Machine Deployment"
// call, and as a return value for "Get Deployment."
type Deployment struct {
    XMLNS          string `xml:"xmlns,attr"`
    XMLNS_I        string `xml:"xmlns:i,attr"`
    Name           string `xml:"Name"`
    DeploymentSlot string `xml:"DeploymentSlot"`
    PrivateID      string `xml:"PrivateID,omitempty"` // Only used for "Get Deployment."
    Status         string `xml:"Status,omitempty"`    // Only used for "Get Deployment."
    Label          string `xml:"Label"`
    URL            string `xml:"Url,omitempty"`           // Only used for "Get Deployment."
    Configuration  string `xml:"Configuration,omitempty"` // Only used for "Get Deployment."
    // Not implemented: RoleInstanceList.  Only used for "Get Deployment."
    UpgradeStatus            string             `xml:"UpgradeStatus,omitempty"`      // Only used for "Get Deployment."
    UpgradeDomainCount       string             `xml:"UpgradeDomainCount,omitempty"` // Only used for "Get Deployment."
    RoleList                 []Role             `xml:"RoleList>Role"`
    SDKVersion               string             `xml:"SdkVersion,omitempty"`      // Only used for "Get Deployment."
    Locked                   string             `xml:"Locked,omitempty"`          // Only used for "Get Deployment."
    RollbackAllowed          string             `xml:"RollbackAllowed,omitempty"` // Only used for "Get Deployment."
    VirtualNetworkName       string             `xml:VirtualNetworkName,omitempty"`
    DNS                      []DnsServer        `xml:"Dns>DnsServers>DnsServer",omitempty`
    ExtendedProperties       []ExtendedProperty `xml:"ExtendedProperties,omitempty"`       // Only used for "Get Deployment."
    PersistentVMDowntimeInfo string             `xml:"PersistentVMDowntimeInfo,omitempty"` // Only used for "Get Deployment."
    // Not implemented: VirtualIPs.  Only used for "Get Deployment."
}

func (c *Deployment) Serialize() (string, error) {
    return toxml(c)
}

// newDeploymentForCreateVMDeployment creates a Deployment object for the
// purpose of passing it to "Create Virtual Machine Deployment."
// You may still want to set the optional DNS attribute.
func NewDeploymentForCreateVMDeployment(name, deploymentSlot, label string, roles []Role, virtualNetworkName string) *Deployment {
    deployment := Deployment{
        XMLNS:              XMLNS,
        XMLNS_I:            XMLNS_I,
        Name:               name,
        DeploymentSlot:     deploymentSlot,
        Label:              base64.StdEncoding.EncodeToString([]byte(label)),
        RoleList:           roles,
        VirtualNetworkName: virtualNetworkName,
    }
    return &deployment
}

const XMLNS = "http://schemas.microsoft.com/windowsazure"
const XMLNS_I = "http://www.w3.org/2001/XMLSchema-instance"

//
// Role Operations bits
//

type RoleOperation struct {
    XMLName       xml.Name
    XMLNS         string `xml:"xmlns,attr"`
    XMLNS_I       string `xml:"xmlns:i,attr"`
    OperationType string `xml:"OperationType"`
}

func newRoleOperation(operationType string) *RoleOperation {
    operation := RoleOperation{
        XMLNS:         XMLNS,
        XMLNS_I:       XMLNS_I,
        OperationType: operationType,
    }
    operation.XMLName.Local = operationType
    return &operation
}

// The Start Role operation starts a virtual machine. 
// http://msdn.microsoft.com/en-us/library/windowsazure/jj157189.aspx
var StartRoleOperation = newRoleOperation("StartRoleOperation")

// The Shutdown Role operation shuts down a virtual machine.
// http://msdn.microsoft.com/en-us/library/windowsazure/jj157195.aspx
var ShutdownRoleOperation = newRoleOperation("ShutdownRoleOperation")

// The Restart role operation restarts a virtual machine.
// http://msdn.microsoft.com/en-us/library/windowsazure/jj157197.aspx
var RestartRoleOperation = newRoleOperation("RestartRoleOperation")

//
// Storage Services bits
//

type ExtendedProperty struct {
    Name  string `xml:"Name"`
    Value string `xml:"Value"`
}

type StorageService struct {
    // List Storage Accounts.
    // See http://msdn.microsoft.com/en-us/library/windowsazure/ee460787.aspx
    Url                   string             `xml:"Url"` // TODO: change the struct field to all-uppercase.
    ServiceName           string             `xml:"ServiceName"`
    Description           string             `xml:"StorageServiceProperties>Description"`
    AffinityGroup         string             `xml:"StorageServiceProperties>AffinityGroup"`
    Label                 string             `xml:"StorageServiceProperties>Label"` // base64
    Status                string             `xml:"StorageServiceProperties>Status"`
    Endpoints             []string           `xml:"StorageServiceProperties>Endpoints>Endpoint"`
    GeoReplicationEnabled string             `xml:"StorageServiceProperties>GeoReplicationEnabled"`
    GeoPrimaryRegion      string             `xml:"StorageServiceProperties>GeoPrimaryRegion"`
    StatusOfPrimary       string             `xml:"StorageServiceProperties>StatusOfPrimary"`
    LastGeoFailoverTime   string             `xml:"StorageServiceProperties>LastGeoFailoverTime"`
    GeoSecondaryRegion    string             `xml:"StorageServiceProperties>GeoSecondaryRegion"`
    StatusOfSecondary     string             `xml:"StorageServiceProperties>StatusOfSecondary"`
    ExtendedProperties    []ExtendedProperty `xml:"StorageServiceProperties>ExtendedProperties>ExtendedProperty"`

    // TODO: Add accessors for non-string data encoded as strings.
}

type StorageServices struct {
    xmlns           string           `xml:"xmlns,attr"`
    StorageServices []StorageService `xml:"StorageService"`
}

func (s *StorageServices) Deserialize(data []byte) error {
    return xml.Unmarshal(data, s)
}

// CreateStorageServiceInput is a request to create a storage account.
// (Azure's "storage services" seem to have been renamed to "storage accounts"
// but the old terminology is still evident in the API).
type CreateStorageServiceInput struct {
    // See http://msdn.microsoft.com/en-us/library/windowsazure/hh264518.aspx
    XMLNS                 string             `xml:"xmlns,attr"`
    ServiceName           string             `xml:"ServiceName"`
    Label                 string             `xml:"Label"`
    Description           string             `xml:"Description,omitempty"`
    Location              string             `xml:"Location"`
    AffinityGroup         string             `xml:"AffinityGroup,omitempty"`
    GeoReplicationEnabled string             `xml:"GeoReplicationEnabled,omitempty"`
    ExtendedProperties    []ExtendedProperty `xml:"ExtendedProperties>ExtendedProperty"`
}

func (c *CreateStorageServiceInput) Serialize() (string, error) {
    return toxml(c)
}

// NewCreateStorageServiceInputWithLocation creates a location-based
// CreateStorageServiceInput, with all required fields filled out.
func NewCreateStorageServiceInputWithLocation(name, label, location string) *CreateStorageServiceInput {
    return &CreateStorageServiceInput{
        XMLNS:       XMLNS,
        ServiceName: name,
        Label:       base64.StdEncoding.EncodeToString([]byte(label)),
        Location:    location,
    }
}

type MetadataItem struct {
    XMLName xml.Name
    Value   string `xml:",chardata"`
}

func (item *MetadataItem) Name() string {
    return item.XMLName.Local
}

type Metadata struct {
    Items []MetadataItem `xml:",any"`
}

type Blob struct {
    Name                  string   `xml:"Name"`
    Snapshot              string   `xml:"Snapshot"`
    Url                   string   `xml:"Url"`
    LastModified          string   `xml:"Properties>Last-Modified"`
    Etag                  string   `xml:"Properties>Etag"`
    ContentLength         string   `xml:"Properties>Content-Length"`
    ContentType           string   `xml:"Properties>Content-Type"`
    BlobSequenceNumber    string   `xml:"Properties>x-ms-blob-sequence-number"`
    BlobType              string   `xml:"Properties>BlobType"`
    LeaseStatus           string   `xml:"Properties>LeaseStatus"`
    LeaseState            string   `xml:"Properties>LeaseState"`
    LeaseDuration         string   `xml:"Properties>LeaseDuration"`
    CopyId                string   `xml:"Properties>CopyId"`
    CopyStatus            string   `xml:"Properties>CopyStatus"`
    CopySource            string   `xml:"Properties>CopySource"`
    CopyProgress          string   `xml:"Properties>CopyProgress"`
    CopyCompletionTime    string   `xml:"Properties>CopyCompletionTime"`
    CopyStatusDescription string   `xml:"Properties>CopyStatusDescription"`
    Metadata              Metadata `xml:"Metadata"`
}

type BlobEnumerationResults struct {
    // http://msdn.microsoft.com/en-us/library/windowsazure/dd135734.aspx
    XMLName       xml.Name `xml:"EnumerationResults"`
    ContainerName string   `xml:"ContainerName,attr"`
    Prefix        string   `xml:"Prefix"`
    Marker        string   `xml:"Marker"`
    MaxResults    string   `xml:"MaxResults"`
    Delimiter     string   `xml:"Delimiter"`
    Blobs         []Blob   `xml:"Blobs>Blob"`
    BlobPrefixes  []string `xml:"Blobs>BlobPrefix>Name"`
}

func (b *BlobEnumerationResults) Deserialize(data []byte) error {
    return xml.Unmarshal(data, b)
}

type StorageAccountKeys struct {
    // See http://msdn.microsoft.com/en-us/library/windowsazure/ee460785.aspx
    XMLName   xml.Name `xml:"StorageService"`
    Url       string   `xml:"Url"` // TODO: change the struct field to all-uppercase.
    Primary   string   `xml:"StorageServiceKeys>Primary"`
    Secondary string   `xml:"StorageServiceKeys>Secondary"`
}

func (s *StorageAccountKeys) Deserialize(data []byte) error {
    return xml.Unmarshal(data, s)
}

type ContainerEnumerationResults struct {
    // See http://msdn.microsoft.com/en-us/library/windowsazure/dd179352.aspx
    XMLName    xml.Name    `xml:"EnumerationResults"`
    Prefix     string      `xml:"Prefix"`
    Marker     string      `xml:"Marker"`
    MaxResults string      `xml:"MaxResults"`
    Containers []Container `xml:"Containers>Container"`
    NextMarker string      `xml:"NextMarker"`
}

func (s *ContainerEnumerationResults) Deserialize(data []byte) error {
    return xml.Unmarshal(data, s)
}

type Container struct {
    XMLName    xml.Name   `xml:"Container"`
    Name       string     `xml:"Name"`
    URL        string     `xml:"URL"`
    Properties Properties `xml:"Properties"`
    Metadata   Metadata   `xml:"Metadata"`
}

type Properties struct {
    LastModified  string `xml:"Last-Modified"`
    Etag          string `xml:"Etag"`
    LeaseStatus   string `xml:"LeaseStatus"`
    LeaseState    string `xml:"LeaseState"`
    LeaseDuration string `xml:"LeaseDuration"`
}

// An enumeration-lite type to define from which list (committed, uncommitted,
// latest) to get a block during a PutBlockList Storage API operation.
type BlockListType string

const (
    BlockListUncommitted BlockListType = "Uncommitted"
    BlockListCommitted   BlockListType = "Committed"
    BlockListLatest      BlockListType = "Latest"
)

// Payload for the PutBlockList operation.
type BlockList struct {
    XMLName xml.Name `xml:"BlockList"`
    Items   []*BlockListItem
}

func (s *BlockList) Serialize() ([]byte, error) {
    return marshalXML(s)
}

// Add a BlockListItem to a BlockList.
func (s *BlockList) Add(blockType BlockListType, blockID string) {
    base64_id := base64.StdEncoding.EncodeToString([]byte(blockID))
    item := NewBlockListItem(blockType, base64_id)
    s.Items = append(s.Items, item)
}

type BlockListItem struct {
    XMLName xml.Name
    BlockID string `xml:",chardata"`
}

// Create a new BlockListItem.
func NewBlockListItem(blockType BlockListType, blockID string) *BlockListItem {
    return &BlockListItem{
        XMLName: xml.Name{Local: string(blockType)},
        BlockID: blockID,
    }
}

func (item *BlockListItem) Type() BlockListType {
    switch BlockListType(item.XMLName.Local) {
    case BlockListUncommitted:
        return BlockListUncommitted
    case BlockListCommitted:
        return BlockListCommitted
    case BlockListLatest:
        return BlockListLatest
    }
    panic(fmt.Errorf("type not recognized: %s", item.XMLName.Local))
}

// GetBlockList result struct.
type Block struct {
    Name string `xml:"Name"`
    Size string `xml:"Size"`
}

type GetBlockList struct {
    XMLName           xml.Name `xml:"BlockList"`
    CommittedBlocks   []Block  `xml:"CommittedBlocks>Block"`
    UncommittedBlocks []Block  `xml:"UncommittedBlocks>Block"`
}

func (g *GetBlockList) Deserialize(data []byte) error {
    return xml.Unmarshal(data, g)
}

//
// Operation Services bits
//

const (
    InProgressOperationStatus = "InProgress"
    SucceededOperationStatus  = "Succeeded"
)

type Operation struct {
    ID             string `xml:"ID"`
    Status         string `xml:"Status"`
    HttpStatusCode int    `xml:"HttpStatusCode"`
    ErrorCode      string `xml:"Error>Code"`
    ErrorMessage   string `xml:"Error>Message"`
}
