// Copyright 2013 Canonical Ltd.  This software is licensed under the
// GNU Lesser General Public License version 3 (see the file COPYING).

package gwacl

import (
    "encoding/base64"
    "encoding/xml"
    "fmt"
    . "launchpad.net/gocheck"
    "launchpad.net/gwacl/dedent"
)

type xmlSuite struct{}

var _ = Suite(&xmlSuite{})

//
// Tests for Marshallers
//

func (suite *xmlSuite) TestLinuxProvisioningConfiguration(c *C) {
    config := makeLinuxProvisioningConfiguration()

    xml, err := config.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <LinuxProvisioningConfiguration>
          <ConfigurationSetType>LinuxProvisioningConfiguration</ConfigurationSetType>
          <HostName>%s</HostName>
          <UserName>%s</UserName>
          <UserPassword>%s</UserPassword>
          <DisableSshPasswordAuthentication>%v</DisableSshPasswordAuthentication>
        </LinuxProvisioningConfiguration>`)
    expected := fmt.Sprintf(template, config.Hostname, config.Username,
        config.Password, config.DisableSshPasswordAuthentication)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestInputEndpoint(c *C) {
    endpoint := makeEndpoint(endpointParams{})

    xml, err := endpoint.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <InputEndpoint>
          <LoadBalancedEndpointSetName>%s</LoadBalancedEndpointSetName>
          <LocalPort>%v</LocalPort>
          <Name>%s</Name>
          <Port>%v</Port>
          <Protocol>%s</Protocol>
        </InputEndpoint>`)
    expected := fmt.Sprintf(template, endpoint.LoadBalancedEndpointSetName,
        endpoint.LocalPort, endpoint.Name, endpoint.Port, endpoint.Protocol)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestOSVirtualHardDisk(c *C) {
    disk := makeOSVirtualHardDisk()

    xml, err := disk.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <OSVirtualHardDisk>
          <HostCaching>%s</HostCaching>
          <DiskLabel>%s</DiskLabel>
          <DiskName>%s</DiskName>
          <MediaLink>%s</MediaLink>
          <SourceImageName>%s</SourceImageName>
        </OSVirtualHardDisk>`)
    expected := fmt.Sprintf(template, disk.HostCaching, disk.DiskLabel,
        disk.DiskName, disk.MediaLink, disk.SourceImageName)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestNetworkConfiguration(c *C) {
    endpoint1 := makeEndpoint(endpointParams{})
    endpoint2 := makeEndpoint(endpointParams{})
    endpoints := []InputEndpoint{*endpoint1, *endpoint2}
    config := &NetworkConfiguration{
        InputEndpoints: endpoints}
    xml, err := config.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <NetworkConfiguration>
          <InputEndpoints>
            <InputEndpoint>
              <LoadBalancedEndpointSetName>%s</LoadBalancedEndpointSetName>
              <LocalPort>%v</LocalPort>
              <Name>%s</Name>
              <Port>%v</Port>
              <Protocol>%s</Protocol>
            </InputEndpoint>
            <InputEndpoint>
              <LoadBalancedEndpointSetName>%s</LoadBalancedEndpointSetName>
              <LocalPort>%v</LocalPort>
              <Name>%s</Name>
              <Port>%v</Port>
              <Protocol>%s</Protocol>
            </InputEndpoint>
          </InputEndpoints>
        </NetworkConfiguration>`)
    expected := fmt.Sprintf(template, endpoint1.LoadBalancedEndpointSetName,
        endpoint1.LocalPort, endpoint1.Name, endpoint1.Port,
        endpoint1.Protocol, endpoint2.LoadBalancedEndpointSetName,
        endpoint2.LocalPort, endpoint2.Name, endpoint2.Port,
        endpoint2.Protocol)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestRole(c *C) {
    role := makeRole()
    config := role.ConfigurationSets[0]

    xml, err := role.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <Role>
          <RoleName>%s</RoleName>
          <RoleType>PersistentVMRole</RoleType>
          <ConfigurationSets>
            <ConfigurationSet>
              <ConfigurationSetType>%s</ConfigurationSetType>
              <HostName>%s</HostName>
              <UserName>%s</UserName>
              <UserPassword>%s</UserPassword>
              <DisableSshPasswordAuthentication>%v</DisableSshPasswordAuthentication>
            </ConfigurationSet>
          </ConfigurationSets>
          <RoleSize>%s</RoleSize>
        </Role>`)
    expected := fmt.Sprintf(template, role.RoleName,
        config.ConfigurationSetType, config.Hostname, config.Username,
        config.Password, config.DisableSshPasswordAuthentication, role.RoleSize)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestDeployment(c *C) {
    deployment := makeDeployment()
    dns := deployment.DNS[0]
    role := deployment.RoleList[0]
    config := role.ConfigurationSets[0]

    xml, err := deployment.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <Deployment xmlns="http://schemas.microsoft.com/windowsazure" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
          <Name>%s</Name>
          <DeploymentSlot>%s</DeploymentSlot>
          <Label>%s</Label>
          <RoleList>
            <Role>
              <RoleName>%s</RoleName>
              <RoleType>PersistentVMRole</RoleType>
              <ConfigurationSets>
                <ConfigurationSet>
                  <ConfigurationSetType>%s</ConfigurationSetType>
                  <HostName>%s</HostName>
                  <UserName>%s</UserName>
                  <UserPassword>%s</UserPassword>
                  <DisableSshPasswordAuthentication>%v</DisableSshPasswordAuthentication>
                </ConfigurationSet>
              </ConfigurationSets>
              <RoleSize>%s</RoleSize>
            </Role>
          </RoleList>
          <VirtualNetworkName>%s</VirtualNetworkName>
          <Dns>
            <DnsServers>
              <DnsServer>
                <Name>%s</Name>
                <Address>%s</Address>
              </DnsServer>
            </DnsServers>
          </Dns>
        </Deployment>`)
    expected := fmt.Sprintf(template, deployment.Name,
        deployment.DeploymentSlot, deployment.Label,
        role.RoleName, config.ConfigurationSetType, config.Hostname,
        config.Username, config.Password,
        config.DisableSshPasswordAuthentication, role.RoleSize,
        deployment.VirtualNetworkName, dns.Name, dns.Address)
    c.Check(xml, Equals, expected)
}

func (suite *xmlSuite) TestNewDeploymentForCreateVMDeployment(c *C) {
    name := "deploymentName"
    deploymentSlot := "staging"
    label := "deploymentLabel"
    OSVirtualHardDisk := NewOSVirtualHardDisk("hostCaching", "diskLabel", "diskName", "mediaLink", "sourceImageName")
    roles := []Role{*NewRole("size", "name", []LinuxProvisioningConfiguration{}, OSVirtualHardDisk)}
    virtualNetworkName := "network"

    deployment := NewDeploymentForCreateVMDeployment(name, deploymentSlot, label, roles, virtualNetworkName)

    c.Check(deployment.XMLNS, Equals, XMLNS)
    c.Check(deployment.XMLNS_I, Equals, XMLNS_I)
    c.Check(deployment.Name, Equals, name)
    c.Check(deployment.DeploymentSlot, Equals, deploymentSlot)
    c.Check(deployment.RoleList, DeepEquals, roles)
    decodedLabel, err := base64.StdEncoding.DecodeString(deployment.Label)
    c.Assert(err, IsNil)
    c.Check(string(decodedLabel), Equals, label)
    c.Check(deployment.VirtualNetworkName, Equals, virtualNetworkName)
}

func (suite *xmlSuite) TestCreateStorageServiceInput(c *C) {
    s := makeCreateStorageServiceInput()
    ext_property := s.ExtendedProperties[0]
    xml, err := s.Serialize()
    c.Assert(err, IsNil)
    template := dedent.Dedent(`
        <CreateStorageServiceInput xmlns="http://schemas.microsoft.com/windowsazure">
          <ServiceName>%s</ServiceName>
          <Label>%s</Label>
          <Description>%s</Description>
          <Location>%s</Location>
          <AffinityGroup>%s</AffinityGroup>
          <GeoReplicationEnabled>%s</GeoReplicationEnabled>
          <ExtendedProperties>
            <ExtendedProperty>
              <Name>%s</Name>
              <Value>%s</Value>
            </ExtendedProperty>
          </ExtendedProperties>
        </CreateStorageServiceInput>`)
    expected := fmt.Sprintf(template, s.ServiceName, s.Label, s.Description,
        s.Location, s.AffinityGroup, s.GeoReplicationEnabled, ext_property.Name,
        ext_property.Value)
    c.Assert(xml, Equals, expected)
}

//
// Tests for Unmarshallers
//

func (suite *xmlSuite) TestStorageServicesUnmarshal(c *C) {
    input_template := `
        <?xml version="1.0" encoding="utf-8"?>
        <StorageServices xmlns="http://schemas.microsoft.com/windowsazure">
          <StorageService>
            <Url>%s</Url>
            <ServiceName>%s</ServiceName>
            <StorageServiceProperties>
              <Description>%s</Description>
              <AffinityGroup>%s</AffinityGroup>
              <Label>%s</Label>
              <Status>%s</Status>
              <Endpoints>
                <Endpoint>%s</Endpoint>
                <Endpoint>%s</Endpoint>
                <Endpoint>%s</Endpoint>
              </Endpoints>
              <GeoReplicationEnabled>%s</GeoReplicationEnabled>
              <GeoPrimaryRegion>%s</GeoPrimaryRegion>
              <StatusOfPrimary>%s</StatusOfPrimary>
              <LastGeoFailoverTime>%s</LastGeoFailoverTime>
              <GeoSecondaryRegion>%s</GeoSecondaryRegion>
              <StatusOfSecondary>%s</StatusOfSecondary>
              <ExtendedProperties>
                <ExtendedProperty>
                  <Name>%s</Name>
                  <Value>%s</Value>
                </ExtendedProperty>
                <ExtendedProperty>
                  <Name>%s</Name>
                  <Value>%s</Value>
                </ExtendedProperty>
              </ExtendedProperties>
            </StorageServiceProperties>
          </StorageService>
        </StorageServices>`
    url := MakeRandomString(10)
    servicename := MakeRandomString(10)
    desc := MakeRandomString(10)
    affinity := MakeRandomString(10)
    label := MakeRandomString(10)
    status := MakeRandomString(10)
    blob_endpoint := MakeRandomString(10)
    queue_endpoint := MakeRandomString(10)
    table_endpoint := MakeRandomString(10)
    geo_repl := BoolToString(MakeRandomBool())
    geo_region := MakeRandomString(10)
    status_primary := MakeRandomString(10)
    failover_time := MakeRandomString(10)
    geo_sec_region := MakeRandomString(10)
    status_sec := MakeRandomString(10)
    p1_name := MakeRandomString(10)
    p1_val := MakeRandomString(10)
    p2_name := MakeRandomString(10)
    p2_val := MakeRandomString(10)

    input := fmt.Sprintf(input_template, url, servicename, desc, affinity,
        label, status, blob_endpoint, queue_endpoint, table_endpoint, geo_repl,
        geo_region, status_primary, failover_time, geo_sec_region, status_sec,
        p1_name, p1_val, p2_name, p2_val)
    data := []byte(input)

    services := &StorageServices{}
    err := services.Deserialize(data)
    c.Assert(err, IsNil)

    c.Check(len(services.StorageServices), Equals, 1)
    s := services.StorageServices[0]

    // Oh jeez, here we go....
    c.Check(s.Url, Equals, url)
    c.Check(s.ServiceName, Equals, servicename)
    c.Check(s.Description, Equals, desc)
    c.Check(s.AffinityGroup, Equals, affinity)
    c.Check(s.Label, Equals, label)
    c.Check(s.Status, Equals, status)
    c.Check(s.GeoReplicationEnabled, Equals, geo_repl)
    c.Check(s.GeoPrimaryRegion, Equals, geo_region)
    c.Check(s.StatusOfPrimary, Equals, status_primary)
    c.Check(s.LastGeoFailoverTime, Equals, failover_time)
    c.Check(s.GeoSecondaryRegion, Equals, geo_sec_region)
    c.Check(s.StatusOfSecondary, Equals, status_sec)

    endpoints := s.Endpoints
    c.Check(len(endpoints), Equals, 3)
    c.Check(endpoints[0], Equals, blob_endpoint)
    c.Check(endpoints[1], Equals, queue_endpoint)
    c.Check(endpoints[2], Equals, table_endpoint)

    properties := s.ExtendedProperties
    c.Check(properties[0].Name, Equals, p1_name)
    c.Check(properties[0].Value, Equals, p1_val)
    c.Check(properties[1].Name, Equals, p2_name)
    c.Check(properties[1].Value, Equals, p2_val)
}

func (suite *xmlSuite) TestBlobEnumerationResuts(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <EnumerationResults ContainerName="http://myaccount.blob.core.windows.net/mycontainer">
          <Prefix>prefix</Prefix>
          <Marker>marker</Marker>
          <MaxResults>maxresults</MaxResults>
          <Delimiter>delimiter</Delimiter>
          <Blobs>
            <Blob>
              <Name>blob-name</Name>
              <Snapshot>snapshot-date-time</Snapshot>
              <Url>blob-address</Url>
              <Properties>
                <Last-Modified>last-modified</Last-Modified>
                <Etag>etag</Etag>
                <Content-Length>size-in-bytes</Content-Length>
                <Content-Type>blob-content-type</Content-Type>
                <Content-Encoding />
                <Content-Language />
                <Content-MD5 />
                <Cache-Control />
                <x-ms-blob-sequence-number>sequence-number</x-ms-blob-sequence-number>
                <BlobType>blobtype</BlobType>
                <LeaseStatus>leasestatus</LeaseStatus>
                <LeaseState>leasestate</LeaseState>
                <LeaseDuration>leasesduration</LeaseDuration>
                <CopyId>id</CopyId>
                <CopyStatus>copystatus</CopyStatus>
                <CopySource>copysource</CopySource>
                <CopyProgress>copyprogress</CopyProgress>
                <CopyCompletionTime>copycompletiontime</CopyCompletionTime>
                <CopyStatusDescription>copydesc</CopyStatusDescription>
              </Properties>
              <Metadata>   
                <MetaName1>metadataname1</MetaName1>
                <MetaName2>metadataname2</MetaName2>
              </Metadata>
            </Blob>
            <BlobPrefix>
              <Name>blob-prefix</Name>
            </BlobPrefix>
          </Blobs>
          <NextMarker />
        </EnumerationResults>`
    data := []byte(input)
    r := &BlobEnumerationResults{}
    err := r.Deserialize(data)
    c.Assert(err, IsNil)
    c.Check(r.ContainerName, Equals, "http://myaccount.blob.core.windows.net/mycontainer")
    c.Check(r.Prefix, Equals, "prefix")
    c.Check(r.Marker, Equals, "marker")
    c.Check(r.Delimiter, Equals, "delimiter")
    b := r.Blobs[0]
    c.Check(b.Name, Equals, "blob-name")
    c.Check(b.Snapshot, Equals, "snapshot-date-time")
    c.Check(b.Url, Equals, "blob-address")
    c.Check(b.LastModified, Equals, "last-modified")
    c.Check(b.Etag, Equals, "etag")
    c.Check(b.ContentLength, Equals, "size-in-bytes")
    c.Check(b.ContentType, Equals, "blob-content-type")
    c.Check(b.BlobSequenceNumber, Equals, "sequence-number")
    c.Check(b.BlobType, Equals, "blobtype")
    c.Check(b.LeaseStatus, Equals, "leasestatus")
    c.Check(b.LeaseState, Equals, "leasestate")
    c.Check(b.LeaseDuration, Equals, "leasesduration")
    c.Check(b.CopyId, Equals, "id")
    c.Check(b.CopyStatus, Equals, "copystatus")
    c.Check(b.CopySource, Equals, "copysource")
    c.Check(b.CopyProgress, Equals, "copyprogress")
    c.Check(b.CopyCompletionTime, Equals, "copycompletiontime")
    c.Check(b.CopyStatusDescription, Equals, "copydesc")
    m1 := b.Metadata.Items[0]
    m2 := b.Metadata.Items[1]
    c.Check(m1.Name(), Equals, "MetaName1")
    c.Check(m1.Value, Equals, "metadataname1")
    c.Check(m2.Name(), Equals, "MetaName2")
    c.Check(m2.Value, Equals, "metadataname2")
    prefix := r.BlobPrefixes[0]
    c.Check(prefix, Equals, "blob-prefix")
}

func (suite *xmlSuite) TestStorageAccountKeysUnmarshal(c *C) {
    template := `
        <?xml version="1.0" encoding="utf-8"?>
          <StorageService xmlns="http://schemas.microsoft.com/windowsazure">
            <Url>%s</Url>
            <StorageServiceKeys>
              <Primary>%s</Primary>
              <Secondary>%s</Secondary>
            </StorageServiceKeys>
          </StorageService>`
    url := MakeRandomString(10)
    key1 := MakeRandomString(10)
    key2 := MakeRandomString(10)
    input := fmt.Sprintf(template, url, key1, key2)
    data := []byte(input)

    keys := &StorageAccountKeys{}
    err := keys.Deserialize(data)
    c.Assert(err, IsNil)
    c.Check(keys.Url, Equals, url)
    c.Check(keys.Primary, Equals, key1)
    c.Check(keys.Secondary, Equals, key2)
}

// Tests for object factory functions.

func (suite *xmlSuite) TestNewRole(c *C) {
    rolesize := MakeRandomString(10)
    rolename := MakeRandomString(10)
    config := makeLinuxProvisioningConfiguration()
    configset := []LinuxProvisioningConfiguration{*config}
    OSVirtualHardDisk := NewOSVirtualHardDisk("hostCaching", "diskLabel", "diskName", "mediaLink", "sourceImageName")

    role := NewRole(rolesize, rolename, configset, OSVirtualHardDisk)
    c.Check(role.RoleSize, Equals, rolesize)
    c.Check(role.RoleName, Equals, rolename)
    c.Check(role.ConfigurationSets, DeepEquals, configset)
    c.Check(role.RoleType, Equals, "PersistentVMRole")
}

func (suite *xmlSuite) TestNewLinuxProvisioningConfiguration(c *C) {
    hostname := MakeRandomString(10)
    username := MakeRandomString(10)
    password := MakeRandomString(10)
    disablessh := MakeRandomBool()

    config := NewLinuxProvisioningConfiguration(
        hostname, username, password, disablessh)
    c.Check(config.Hostname, Equals, hostname)
    c.Check(config.Username, Equals, username)
    c.Check(config.Password, Equals, password)
    c.Check(config.DisableSshPasswordAuthentication, Equals, disablessh)
    c.Check(config.ConfigurationSetType, Equals, "LinuxProvisioningConfiguration")
}

func (suite *xmlSuite) TestNewOSVirtualHardDisk(c *C) {
    var hostcaching HostCachingType = "ReadWrite"
    disklabel := MakeRandomString(10)
    diskname := MakeRandomString(10)
    MediaLink := MakeRandomString(10)
    SourceImageName := MakeRandomString(10)

    disk := NewOSVirtualHardDisk(
        hostcaching, disklabel, diskname, MediaLink, SourceImageName)
    c.Check(disk.HostCaching, Equals, string(hostcaching))
    c.Check(disk.DiskLabel, Equals, disklabel)
    c.Check(disk.DiskName, Equals, diskname)
    c.Check(disk.MediaLink, Equals, MediaLink)
    c.Check(disk.SourceImageName, Equals, SourceImageName)
}

// Properties XML subtree for ListContainers Storage API call.
func (suite *xmlSuite) TestProperties(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <Properties>
          <Last-Modified>date/time-value</Last-Modified>
          <Etag>etag-value</Etag>
          <LeaseStatus>lease-status-value</LeaseStatus>
          <LeaseState>lease-state-value</LeaseState>
          <LeaseDuration>lease-duration-value</LeaseDuration>
        </Properties>`
    observed := &Properties{}
    err := xml.Unmarshal([]byte(input), observed)
    c.Assert(err, IsNil)

    expected := &Properties{
        LastModified:  "date/time-value",
        Etag:          "etag-value",
        LeaseStatus:   "lease-status-value",
        LeaseState:    "lease-state-value",
        LeaseDuration: "lease-duration-value",
    }

    c.Assert(observed, DeepEquals, expected)
}

// Metadata XML subtree for ListContainers Storage API call.
func (suite *xmlSuite) TestMetadata(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <Metadata>
          <metadata-name>metadata-value</metadata-name>
        </Metadata>`
    observed := &Metadata{}
    err := xml.Unmarshal([]byte(input), observed)
    c.Assert(err, IsNil)

    expected := &Metadata{
        Items: []MetadataItem{
            {
                XMLName: xml.Name{Local: "metadata-name"},
                Value:   "metadata-value",
            },
        },
    }

    c.Assert(observed, DeepEquals, expected)
}

// Container XML subtree for ListContainers Storage API call.
func (suite *xmlSuite) TestContainer(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <Container>
          <Name>name-value</Name>
          <URL>url-value</URL>
          <Properties>
            <Last-Modified>date/time-value</Last-Modified>
            <Etag>etag-value</Etag>
            <LeaseStatus>lease-status-value</LeaseStatus>
            <LeaseState>lease-state-value</LeaseState>
            <LeaseDuration>lease-duration-value</LeaseDuration>
          </Properties>
          <Metadata>
            <metadata-name>metadata-value</metadata-name>
          </Metadata>
        </Container>`
    observed := &Container{}
    err := xml.Unmarshal([]byte(input), observed)
    c.Assert(err, IsNil)

    expected := &Container{
        XMLName: xml.Name{Local: "Container"},
        Name:    "name-value",
        URL:     "url-value",
        Properties: Properties{
            LastModified:  "date/time-value",
            Etag:          "etag-value",
            LeaseStatus:   "lease-status-value",
            LeaseState:    "lease-state-value",
            LeaseDuration: "lease-duration-value",
        },
        Metadata: Metadata{
            Items: []MetadataItem{
                {
                    XMLName: xml.Name{Local: "metadata-name"},
                    Value:   "metadata-value",
                },
            },
        },
    }

    c.Assert(observed, DeepEquals, expected)
}

// EnumerationResults XML tree for ListContainers Storage API call.
func (suite *xmlSuite) TestContainerEnumerationResults(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <EnumerationResults AccountName="http://myaccount.blob.core.windows.net">
          <Prefix>prefix-value</Prefix>
          <Marker>marker-value</Marker>
          <MaxResults>max-results-value</MaxResults>
          <Containers>
            <Container>
              <Name>name-value</Name>
              <URL>url-value</URL>
              <Properties>
                <Last-Modified>date/time-value</Last-Modified>
                <Etag>etag-value</Etag>
                <LeaseStatus>lease-status-value</LeaseStatus>
                <LeaseState>lease-state-value</LeaseState>
                <LeaseDuration>lease-duration-value</LeaseDuration>
              </Properties>
              <Metadata>
                <metadata-name>metadata-value</metadata-name>
              </Metadata>
            </Container>
          </Containers>
          <NextMarker>next-marker-value</NextMarker>
        </EnumerationResults>`
    observed := &ContainerEnumerationResults{}
    err := observed.Deserialize([]byte(input))
    c.Assert(err, IsNil)

    expected := &ContainerEnumerationResults{
        XMLName:    xml.Name{Local: "EnumerationResults"},
        Prefix:     "prefix-value",
        Marker:     "marker-value",
        MaxResults: "max-results-value",
        Containers: []Container{
            {
                XMLName: xml.Name{Local: "Container"},
                Name:    "name-value",
                URL:     "url-value",
                Properties: Properties{
                    LastModified:  "date/time-value",
                    Etag:          "etag-value",
                    LeaseStatus:   "lease-status-value",
                    LeaseState:    "lease-state-value",
                    LeaseDuration: "lease-duration-value",
                },
                Metadata: Metadata{
                    Items: []MetadataItem{
                        {
                            XMLName: xml.Name{Local: "metadata-name"},
                            Value:   "metadata-value",
                        },
                    },
                },
            },
        },
        NextMarker: "next-marker-value",
    }

    c.Assert(observed, DeepEquals, expected)
    c.Assert(observed.Containers[0].Metadata.Items[0].Name(), Equals, "metadata-name")
}

func (suite *xmlSuite) TestNewCreateHostedServiceWithLocation(c *C) {
    serviceName := "serviceName"
    label := "label"
    location := "location"
    createdHostedService := NewCreateHostedServiceWithLocation(serviceName, label, location)
    base64label := base64.StdEncoding.EncodeToString([]byte(label))
    c.Check(createdHostedService.ServiceName, DeepEquals, serviceName)
    c.Check(createdHostedService.Label, DeepEquals, base64label)
    c.Check(createdHostedService.Location, DeepEquals, location)
}

func (suite *xmlSuite) TestNewCreateStorageServiceInputWithLocation(c *C) {
    cssi := NewCreateStorageServiceInputWithLocation("name", "label", "location")
    c.Check(cssi.XMLNS, Equals, XMLNS)
    c.Check(cssi.ServiceName, Equals, "name")
    c.Check(cssi.Label, Equals, base64.StdEncoding.EncodeToString([]byte("label")))
    c.Check(cssi.Location, Equals, "location")
}

func (suite *xmlSuite) TestBlockListSerialize(c *C) {
    block_list := &BlockList{
        XMLName: xml.Name{Local: "BlockList"},
    }
    block_list.Add(BlockListCommitted, "first-base64-encoded-block-id")
    block_list.Add(BlockListUncommitted, "second-base64-encoded-block-id")
    block_list.Add(BlockListLatest, "third-base64-encoded-block-id")
    observed, err := block_list.Serialize()
    c.Assert(err, IsNil)
    expected := dedent.Dedent(`
        <BlockList>
          <Committed>Zmlyc3QtYmFzZTY0LWVuY29kZWQtYmxvY2staWQ=</Committed>
          <Uncommitted>c2Vjb25kLWJhc2U2NC1lbmNvZGVkLWJsb2NrLWlk</Uncommitted>
          <Latest>dGhpcmQtYmFzZTY0LWVuY29kZWQtYmxvY2staWQ=</Latest>
        </BlockList>`)
    c.Assert(string(observed), Equals, expected)
}

func (suite *xmlSuite) TestGetBlockListDeserialize(c *C) {
    input := `
        <?xml version="1.0" encoding="utf-8"?>
        <BlockList>
          <CommittedBlocks>
            <Block>
              <Name>BlockId001</Name>
              <Size>4194304</Size>
            </Block>
          </CommittedBlocks>
          <UncommittedBlocks>
            <Block>
              <Name>BlockId002</Name>
              <Size>1024</Size>
            </Block>
          </UncommittedBlocks>
        </BlockList>`
    observed := GetBlockList{}
    err := observed.Deserialize([]byte(input))
    c.Assert(err, IsNil)
    expected := GetBlockList{
        XMLName: xml.Name{Local: "BlockList"},
        CommittedBlocks: []Block{
            {
                Name: "BlockId001",
                Size: "4194304"},
        },
        UncommittedBlocks: []Block{
            {
                Name: "BlockId002",
                Size: "1024"},
        },
    }
    c.Check(observed, DeepEquals, expected)
}

func makeOperationXML(operationType string) string {
    XMLTemplate := dedent.Dedent(`
        <%s xmlns="http://schemas.microsoft.com/windowsazure" xmlns:i="http://www.w3.org/2001/XMLSchema-instance">
          <OperationType>%s</OperationType>
        </%s>`)
    return fmt.Sprintf(XMLTemplate, operationType, operationType, operationType)
}

func (suite *managementAPISuite) TestStartRoleOperation(c *C) {
    expectedXML := makeOperationXML("StartRoleOperation")
    xml, err := marshalXML(StartRoleOperation)
    c.Assert(err, IsNil)
    c.Check(string(xml), Equals, expectedXML)
}

func (suite *managementAPISuite) TestRestartRoleOperation(c *C) {
    expectedXML := makeOperationXML("RestartRoleOperation")
    xml, err := marshalXML(RestartRoleOperation)
    c.Assert(err, IsNil)
    c.Check(string(xml), Equals, expectedXML)
}

func (suite *managementAPISuite) TestShutdownRoleOperation(c *C) {
    expectedXML := makeOperationXML("ShutdownRoleOperation")
    xml, err := marshalXML(ShutdownRoleOperation)
    c.Assert(err, IsNil)
    c.Check(string(xml), Equals, expectedXML)
}
